from distutils.version import LooseVersion
from os import rename
from typing import Union

import dask.array
import six
import xarray
from xarray.backends.netCDF4_ import NetCDF4DataStore
from pathlib import Path

try:
    from enstools.encoding.api import FilterEncodingForXarray, check_dataset_filters_availability
    encoding_available = True
except ModuleNotFoundError:
    encoding_available = False

from .file_type import get_file_type


def write(ds: Union[xarray.Dataset, xarray.DataArray],
          file_path: Union[str, Path],
          file_format: Union[str, None] = None,
          compression: Union[str, dict, None] = None,
          compute: bool =True,
          engine: str ="h5netcdf",
          format: str ="NETCDF4",
          ):
    """
    write a xarray dataset to a file

    Parameters
    ----------
    ds : xarray.Dataset
            the dataset to store

    file_path : string or Path
            the file to create

    file_format : {'NC'}
            string indicating the format to use. if not specified, the file extension if used.
            
    compression : string
            Used to specify the compression mode and optionally additional arguments.
            The parameter follows the rules defined in enstools-encoding.
            (https://gitlab.physik.uni-muenchen.de/w2w/enstools-encoding.git)
            
            To apply lossless compression we can just use:
                "lossless"
            Or we can select the backend and the compression level using the following syntax:
                "lossless,backend,compression_level"
            The backend can be one of:
                    'blosclz' 
                    'lz4' (default)
                    'lz4hc'
                    'snappy'
                    'zlib'
                    'zstd'
            and the compression level must be an integer from 1 to 9 (default is 9).
            Few examples:
                "lossless,zstd,4"
                "lossless,lz4,9"
                "lossless,snappy,1"
            Using "lossless" without additional arguments would be equivalent to "lossless,lz4,9"
            
            For lossy compression, we might be able to pass more arguments:
                "lossy"
            The lossy compressors available will be:
                'zfp'
                'sz'
            For ZFP we have different compression methods available:
                'rate'
                'accuracy'
                'precision'
            For SZ we have also different compression methods available:
                'abs'
                'rel'
                'pw_rel'
            Each one of this methods require an additional parameter: the rate, the precision or the accuracy.
            The examples would look like:
                'lossy,zfp,accuracy,0.2'
                'lossy,zfp,rate,4'
            There are also few features that target datasets with multiple variables.
            One can write a different specification for different variables by using a
            list of space separated specifications:

            'var1:lossy,zfp,rate,4.0 var2:lossy,sz,abs,0.1'

            For more details see the corresponding documentation.

            Another option would be to pass the path to a YAML configuration file as argument.
    compute : bool 
            Dask delayed feature. Set to true to delay the file writing.
    """

    file_path = Path(file_path).resolve()

    # if ds is a DataVariable instead of a Dataset, then convert it
    if isinstance(ds, xarray.DataArray):
        ds = ds.to_dataset()

    # select the type of file to create
    valid_formats = ["NC"]
    if file_format is not None:
        selected_format = file_format
    else:
        selected_format = get_file_type(file_path.name, only_extension=True)
    if selected_format not in valid_formats:
        raise ValueError("the format '%s' is not (yet) supported!" % selected_format)

    if compression is not None:
        help_message = "To use the compression argument please install enstools-encoding:\n" \
                       "pip install @ git+https://github.com/wavestoweather/enstools-encoding.git"
        assert encoding_available, ModuleNotFoundError(help_message)
        # If a compression variable has been provided, define the proper encoding for HDF5 filters:
        dataset_encoding = FilterEncodingForXarray(dataset=ds, compression=compression)
        dataset_encoding.add_metadata()
    else:
        dataset_encoding = None


    if format == "NETCDF4_CLASSIC" and engine == "h5netcdf":
        raise AssertionError(
            f"enstools.io.write:: Format {format} and engine {engine} are not compatible."
            f"If format {format} is needed, the suggested engine is NETCDF4")

    if engine == "netcdf4":
        if compression not in [None, "default"]:
            logging.warning("Using netcdf4 engine. Setting encoding to None")
            dataset_encoding = None

    # write a netcdf file
    if selected_format == "NC":
        # We can do the trick of changing the name to filename.tmp and changing it back after the process is completed
        # but only if we execute the task here ( i.e. compute==True).
        if compute:
            final_filename = file_path
            file_path = f"{file_path}.tmp"
        task = ds.to_netcdf(file_path, engine=engine, encoding=dataset_encoding, compute=compute, format=format)
        if compute:
            rename(file_path, final_filename)
        return task

        """
        #Old workaround (need to be sure that its not needed anymore)
        # are there dask arrays in the dataset? if so, write the file variable by variable
        if not has_dask_arrays(ds):
            ds.to_netcdf(filename)
        else:
            __to_netcdf(ds, filename)
        """


def __to_netcdf(ds, filename):
    """
    workaround for bug if dask arrays

    Parameters
    ----------
    ds : xarray.Dataset
            dataset to store

    filename : string
            name of the new file
    """
    print("Its using __to_netcdf!")
    # store the complete file without the dask arrays
    ds_copy = ds.copy(deep=False)
    dask_variables = []
    for varname, var in six.iteritems(ds.variables):
        if isinstance(var.data, dask.array.core.Array):
            dask_variables.append(varname)
    for one_var in dask_variables:
        del ds_copy[one_var]
    ds_copy.to_netcdf(filename, format="NETCDF4_CLASSIC")

    # open it again and add the dask arrays
    if LooseVersion(xarray.__version__) > LooseVersion('0.9.6'):
        nc = NetCDF4DataStore.open(filename, "a")
    else:
        nc = NetCDF4DataStore(filename, "a")

    # add the variables
    for varname, var in six.iteritems(ds.variables):
        if isinstance(var.data, dask.array.core.Array):
            nc.store({varname: var.compute()}, {})

    # close the file
    nc.close()



import numpy as np
from enstools.core import check_arguments


def threshold_data(arr, threshold):
    """
    set elements of arr to zero if <= thres

    Parameters
    ----------
    arr : xarray.DataArray or np.ndarray
            input array

    threshold : float
            threshold below which values should be set to zero

    Returns
    -------

    """
    # set elements of arr to zero if <= thres
    return np.where(arr <= threshold, 0, arr)


@check_arguments(shape={"obs": (0, 0),
                        "fct": "obs",
                        "xd_o": "obs",
                        "yd_o": "obs",
                        "lse_o": "obs",
                        "xd_f": "obs",
                        "yd_f": "obs",
                        "lse_f": "obs"})
def calc_das(obs, fct,
             xd_o, yd_o, lse_o,
             xd_f, yd_f, lse_f,
             dis_max,
             threshold=1., x1=0, y1=0, x2=0, y2=0):
    """
    Calculate DAS (Displacement and Amplitude Score)

    Parameters
    ----------
    obs : xarray.DataArray or np.ndarray
            observation

    fct : xarray.DataArray or np.ndarray
            forecast

    xd_o : xarray.DataArray or np.ndarray
            x-displacement in observation space

    yd_o : xarray.DataArray or np.ndarray
            y-displacement in observation space

    lse_o : xarray.DataArray or np.ndarray
            least-square errors in observation space

    xd_f : xarray.DataArray or np.ndarray
            x-displacement in forecast space

    yd_f : xarray.DataArray or np.ndarray
            y-displacement in forecast space

    lse_f : xarray.DataArray or np.ndarray
            least-square errors in forecast space

    dis_max : float
            maximum search distance

    threshold : float
            threshold value (default: 1.)

    x1 : int
            cut of region, left end (default: 0)

    y1 : int
            cut of region, upper end (default: 0)

    x2 : int
            cut of region, right end (default: 0)

    y2 : int
            cut of region, lower end (default: 0)

    Returns
    -------
    tuple:
            das, dis, amp, rms_obs
    """

    # cut off ends
    if x1 or x2 or y1 or y2:
        obs = obs[y1:y2, x1:x2]
        fct = fct[y1:y2, x1:x2]
        xd_o = xd_o[y1:y2, x1:x2]
        yd_o = yd_o[y1:y2, x1:x2]
        lse_o = lse_o[y1:y2, x1:x2]
        xd_f = xd_f[y1:y2, x1:x2]
        yd_f = yd_f[y1:y2, x1:x2]
        lse_f = lse_f[y1:y2, x1:x2]

    dis_o = np.sqrt(xd_o ** 2 + yd_o ** 2)
    dis_f = np.sqrt(xd_f ** 2 + yd_f ** 2)

    if threshold is not None:
        cond_o = np.where((obs > threshold))
        cond_f = np.where((fct > threshold))
    else:
        cond_o = np.empty(obs.shape, dtype=bool)
        cond_o[:] = True
        cond_f = cond_o
    n_obs = np.size(obs[cond_o])
    n_fct = np.size(fct[cond_f])

    dis_o_mean = amp_o_mean = dis_f_mean = amp_f_mean = rms_obs = 0.

    if n_obs > 0:
        rms_obs = np.sqrt(np.mean(obs[cond_o] ** 2))  # root mean square of observation
        dis_o_mean = np.mean(dis_o[cond_o])
        amp_o_mean = np.sqrt(np.mean(lse_o[cond_o]))  # root mean square error
    if n_fct > 0:
        dis_f_mean = np.mean(dis_f[cond_f])
        amp_f_mean = np.sqrt(np.mean(lse_f[cond_f]))  # root mean square error

    if n_obs or n_fct:
        # Displacement Error
        dis = (n_obs * dis_o_mean + n_fct * dis_f_mean) / (n_obs + n_fct)
        # Amplitude Error
        amp = (n_obs * amp_o_mean + n_fct * amp_f_mean) / (n_obs + n_fct)

        # DAS - Displacement and amplitude score
        das = dis / dis_max + amp / rms_obs

        return das, dis / dis_max, amp / rms_obs, rms_obs

    else:
        return np.nan, np.nan, np.nan, np.nan



import os
import logging
import bz2
import six
import xarray
import numpy as np
from numba import jit
import dask
from shutil import copyfileobj


if six.PY3:
    from urllib.request import urlretrieve, urlopen
else:
    from urllib import urlretrieve


def download(url, destination, uncompress=True):
    """
    Download a file from the web

    Parameters
    ----------
    url : string
            the web address

    destination : string
            path to store the file. The file will only be downloaded once.

    uncompress : bool
            if True and if the name ends with '.bz2', it will
            automatically be uncompressed.
    """
    if destination.endswith(".bz2") and uncompress:
        destination_intern = destination[:-4]
    else:
        destination_intern = destination

    if os.path.exists(destination_intern):
        logging.warning("file not downloaded because it is already present: %s" % url)
        return destination_intern

    # download
    logging.info("downloading %s ..." % os.path.basename(destination))
    logging.debug("from: %s" % url)
    fn, hd = urlretrieve(url, destination)

    # uncompress
    if destination.endswith(".bz2") and uncompress:
        # uncompress
        bfile = bz2.BZ2File(destination)
        dfile = open(destination_intern, "wb")
        dfile.write(bfile.read())
        dfile.close()
        bfile.close()
        # delete compressed
        os.remove(destination)

    # return the name if the downloaded file
    return destination_intern


@jit(["b1(f4[:],f4[:],f4,f4)", "b1(f8[:],f8[:],f8,f8)"], nopython=True)
def point_in_polygon(polyx, polyy, testx, testy):
    """
    check whether or not a given coordinate is inside or outside of a polygon

    Parameters
    ----------
    polyx : np.ndarray
            x-coordinates of the polygon

    polyy : np.ndarray
            y-coordinates of the polygon

    testx : float
            x-coordinate of the point

    testy : float
            y-coordinate of the point

    Returns
    -------
    bool
            True, if the point is inside of the polygon
    """
    res = False
    j = polyx.shape[0] - 1
    for i in range(polyx.shape[0]):
        if ((polyy[i] > testy) != (polyy[j] > testy)) \
                and (testx < (polyx[j] - polyx[i]) * (testy - polyy[i]) / (polyy[j] - polyy[i]) + polyx[i]):
            res = not res
        j = i
    return res


@jit(nopython=True)
def distance(lat1, lat2, lon1, lon2, radius=6371229.0, input_in_radian=True):
    """
    distance between two points on a globe.

    Parameters
    ----------
    lat1:
            latitude of first point in radian or degree

    lat2:
            latitude of second point in radian or degree

    lon1:
            longitude of first point in radian or degree

    lon2:
            logitude of second point in radian or degree

    radius:
            radius of the globe in m.

    input_in_radian: bool
            if True, the input is assumed to be in radian, otherwise in degree

    Returns
    -------
    distance in m
    """
    if not input_in_radian:
        lat1 = np.deg2rad(lat1)
        lat2 = np.deg2rad(lat2)
        lon1 = np.deg2rad(lon1)
        lon2 = np.deg2rad(lon2)
    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = np.sin(dlat / 2)**2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon / 2)**2
    c = 2 * np.arctan2(np.sqrt(a), np.sqrt(1 - a))
    return radius * c


def spherical2cartesian(lon, lat, radius=6371229.0):
    """
    calculate cartesian 3d coordinates from spherical coordinates

    Parameters
    ----------
    lon: np.array
            longitude in radian

    lat: np.array
            latitude in radian

    radius: float
            radius of the earth.

    Returns
    -------
    np.array (n, 3)
            cartesian coordinates (x, y, z)
    """
    # create result array
    result = np.empty((lon.shape[0], 3))
    _lat = lat - np.pi / 2.0
    _lon = lon + np.pi
    result[:, 0] = radius * np.sin(_lat) * np.cos(_lon)
    result[:, 1] = radius * np.sin(_lat) * np.sin(_lon)
    result[:, 2] = radius * np.cos(_lat)
    return result


def generate_coordinates(res, grid="regular", lon_range=[-180, 180], lat_range=[-90, 90], unit="degrees"):
    """
    Generate grid coordinates for different types of grids. Currently only regular grids are implemented.

    Parameters
    ----------
    res : float
            resolution of the grid to generate

    grid : {'regular'}
            type of grid to generate. Currently only regular grids are supported
    
    lon_range : list or tuple
            range of longitudes the new grid should cover. Default: -180 to 180

    lat_range : list or tuple
            range of latitudes the new grid should cover. Default: -90 to 90

    unit: {'degrees', 'radians'}

    Returns
    -------
    lon, lat : xarray.DataArray
            tuple of coordinate arrays

    Examples
    --------
    >>> lon, lat = generate_coordinates(20.0, "regular")
    >>> lon
    <xarray.DataArray 'lon' (lon: 18)>
    array([-180., -160., -140., -120., -100.,  -80.,  -60.,  -40.,  -20.,    0.,
             20.,   40.,   60.,   80.,  100.,  120.,  140.,  160.])
    Dimensions without coordinates: lon
    Attributes:
        units:    degrees_east
    >>> lat
    <xarray.DataArray 'lat' (lat: 9)>
    array([-80., -60., -40., -20.,   0.,  20.,  40.,  60.,  80.])
    Dimensions without coordinates: lat
    Attributes:
        units:    degrees_north
    """
    if grid == "regular":
        lon = xarray.DataArray(np.arange(lon_range[0], lon_range[1], res), dims=("lon",), name="lon", attrs={"units": "degrees_east"})
        lat = xarray.DataArray(np.arange(lat_range[0]+res/2.0, lat_range[1], res), dims=("lat",), name="lat", attrs={"units": "degrees_north"})
        if unit == "degrees":
            lon.attrs["units"] = "degrees_east"
            lat.attrs["units"] = "degrees_north"
        elif unit == "radians":
            lon.attrs["units"] = "radians"
            lat.attrs["units"] = "radians"
            lon *= np.pi / 180.0
            lat *= np.pi / 180.0
        else:
            raise ValueError(f"unsupported unit: {unit}")
    else:
        raise ValueError("unsupported grid type: '%s'" % grid)

    return lon, lat


def swapaxis(array, a1, a2):
    """
    move an axis of an array from one position to another. This function belongs to the numpy library but is not
    implemented for xarray.DataArray

    Parameters
    ----------
    array : xarray.DataArray or np.ndarray
            array to manipulate

    a1 : int
            first axis

    a2 : int
            second axis

    Returns
    -------
    xarray.DataArray or np.ndarray
            array with swap axis
    """
    dims = np.arange(array.ndim)
    dims[a1] = a2
    dims[a2] = a1
    dims = tuple(dims)
    if isinstance(array, xarray.DataArray):
        dims = tuple(map(lambda x:array.dims[x], dims))
    return array.transpose(*dims)


def has_ensemble_dim(ds):
    """
    check whether or not a dataset or xarray has already an ensemble dimension

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray

    Returns
    -------
    bool
    """
    return get_ensemble_dim(ds) is not None


def get_ensemble_dim(ds):
    """
    get the name of the ensemble dimension from a dataset or array

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray

    Returns
    -------
    str or None :
            if no ensemble dimension was found, None is returned.
    """
    ens_names = ["ens", "ensemble", "member", "members"]
    for ens_name in ens_names:
        if ens_name in ds.dims:
            logging.debug("get_ensemble_dim: found name '%s'" % ens_name)
            return ens_name
    return None


def set_ensemble_member(ds, member):
    """
    set the number of the ensemble member. The Dataset is modified inplace.

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray
            Array or Dataset with ensemble dimension and only one member
    """
    # get the name of the ensemble dimension
    ens_dim = get_ensemble_dim(ds)

    # if we dont have one, add one
    if ens_dim is None:
        add_ensemble_dim(ds, member)
    else:
        if ds.coords[ens_dim].size == 1:
            ds.coords[ens_dim] = [member]
        else:
            logging.debug("set_ensemble_member: ds has more than one member. Doing nothing!")


def get_time_dim(ds):
    """
    get the name of the time dimension from a dataset or array

    Parameters
    ----------
    ds : xarray.Dataset or xarray.DataArray

    Returns
    -------
    str or None :
            if no ensemble dimension was found, None is returned.
    """
    time_names = ["time", "Time", "times", "Times"]
    for time_name in time_names:
        if time_name in ds.dims:
            logging.debug("get_time_dim: found name '%s'" % time_name)
            return time_name
    return None


def has_dask_arrays(dataset):
    """
    check whether or not a dataset contains dask arrays

    Parameters
    ----------
    dataset : xarray.Dataset

    Returns
    -------
    bool :
            True if dask arrays are found
    """
    has_dask = False
    for varname, var in six.iteritems(dataset.variables):
        if isinstance(var.data, dask.array.core.Array):
            has_dask = True
            break
    return has_dask


def add_ensemble_dim(ds, member, inplace=True):
    """
    create an ensemble dimension with in the dataset

    Parameters
    ----------
    ds : xarray.Dataset

    member : int
            number of the ensemble member

    inplace : bool
            modify the dataset directly?

    Returns
    -------
    xarray.Dataset:
            A copy of the dataset expanded by the ensemble dimension or the expanded dataset
            if the expansion was done inplace.
    """
    # create a lazy copy of the dataset
    if inplace:
        new_ds = ds
    else:
        new_ds = ds.copy()

    # loop over all data variables.
    # those with time dimension are extended behind the time dimension, others at the front
    for one_name, one_var in six.iteritems(ds.data_vars):
        if is_additional_coordinate_variable(one_var):
            continue
        if len(one_var.dims) > 0 and one_var.dims[0] == "time":
            new_ds[one_name] = one_var.expand_dims("ens", 1)
        else:
            new_ds[one_name] = one_var.expand_dims("ens")
    new_ds.coords["ens"] = [member]

    # remove the ensemble_member attribute if present. it if only intented for datasets without ensemble dimension
    if "ensemble_member" in new_ds.attrs:
        del new_ds.attrs["ensemble_member"]
    return new_ds


def is_additional_coordinate_variable(var):
    """
    check if a variable belongs to a list of known constant coordinate variables

    Parameters
    ----------
    var : xarray.DataArray

    Returns
    -------
    bool :
           True = variable is not different for ensemble members
    """

    # list of excluded variables from different models
    excluded = {
        # COSMO variables
        "time_bnds": ("time", "bnds"),
        "slonu": ("rlat", "srlon"),
        "slatu": ("rlat", "srlon"),
        "slonv": ("srlat", "rlon"),
        "slatv": ("srlat", "rlon"),
        "vcoord": ("level1",),
        "soil1_bnds": ("soil1", "bnds"),
        "rotated_pole": (),
        "height_2m": (),
        "height_10m": (),
        "height_toa": (),
        "wbt_13c": (),
    }

    # variable is in excluded list?
    if var.name in excluded and excluded[var.name] == var.dims:
        return True
    else:
        return False


def first_element(array):
    """
    returns the first element of an array or the value of an scalar.

    Parameters
    ----------
    array : xarray.DataArray
            array with 0 or more dimensions.

    Returns
    -------
    float :
            first value.
    """
    if array.size > 1:
        return float(array[0])
    else:
        return float(array)


def count_ge(array, th=0):
    """
    count the number of values above a given threshold (>=)
    Parameters
    ----------
    array : xarray.DataArray and numpy.ndarray
            an array with arbitrary number of dimensions

    th : float
            threshold to test the array for.

    Returns
    -------
    int :
            number of values greater then or equal the threshold.
    """
    if type(array) == xarray.DataArray:
        return __count_ge(array.data, th)
    else:
        return __count_ge(array, th)


@jit(nopython=True)
def __count_ge(array, th):
    """
    numba implementation of count_ge
    """
    result = 0
    for i in range(array.size):
        if array.flat[i] >= th:
            result += 1
    return result


def concat(files, out_filename):
    """
    Concatenates multiple files to one.
    Parameters
    ----------
    files: list
        The list of the files to concat
    out_filename: str
        The name (with destination) of the merged file

    Returns
    -------

    """
    with open(out_filename, "wb") as out:
        for filename in files:
            with open(filename, "rb") as file:
                copyfileobj(file, out)


def bytes2human(n, format='%(value).1f %(symbol)s', symbols='iec'):
    """
    Convert n bytes into a human readable string based on format.
    symbols can be either "customary", "customary_ext", "iec" or "iec_ext",
    see: http://goo.gl/kTQMs
    """
    SYMBOLS = {
        'customary': ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y'),
        'customary_ext': ('byte', 'kilo', 'mega', 'giga', 'tera', 'peta', 'exa', 'zetta', 'iotta'),
        'customary_with_B': ('B', 'KB', 'MB', 'GB', 'TB', 'PB', 'EB', 'ZB', 'YB'),
        'iec': ('Bi', 'Ki', 'Mi', 'Gi', 'Ti', 'Pi', 'Ei', 'Zi', 'Yi'),
        'iec_ext': ('byte', 'kibi', 'mebi', 'gibi', 'tebi', 'pebi', 'exbi', 'zebi', 'yobi'),
        'iec_with_B': ('Bi', 'KiB', 'MiB', 'GiB', 'TiB', 'PiB', 'EiB', 'ZiB', 'YiB'),
    }

    n = int(n)
    if n < 0:
        raise ValueError("n < 0")
    if "iec" in symbols:
        base = 1024
    else:
        base = 1000
    symbols = SYMBOLS[symbols]
    prefix = {}
    for i, s in enumerate(symbols):
        prefix[s] = base ** i
    for symbol in reversed(symbols):
        if n >= prefix[symbol]:
            value = float(n) / prefix[symbol]
            return format % locals()
    return format % dict(symbol=symbols[0], value=n)


@jit(nopython=True)
def isfill(value, fill_value):
    """
    check if a given value is the fill value. Work for numbers and NaN.

    Parameters
    ----------
    value: float
    fill_value: float

    Returns
    -------
    bool:
        True if the value is the fill_value
    """
    if np.isnan(fill_value):
        return np.isnan(value)
    else:
        return value == fill_value



import numpy as np
import xarray

from .normalized_root_mean_square_error import mean_square_error


def peak_signal_to_noise_ratio(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    """
    Description:
        It returns the Peak Sign to Noise Ratio.
        With images, it is usually defined as:
        20*log10(max_value)-10*log10(mse)
        with all the values being between 0 and max_value.
        However, for real values I found this definition:
        20*log10(range)-10*log10(mse)

    Functions
    :param reference:
    :param target:
    :return:
    """

    non_temporal_dimensions = [d for d in reference.dims if d != "time"]
    # Compute range from the reference file
    ref_min, ref_max = reference.min(dim=non_temporal_dimensions), reference.max(dim=non_temporal_dimensions)
    ref_range = ref_max - ref_min

    return 20 * np.log10(ref_range) - 10 * np.log10(mean_square_error(reference, target))



from typing import Union

import numpy
import xarray
from scipy.stats import ks_2samp
from .fix_attributes import fix_attributes
from enstools.core.errors import EnstoolsError


def kolmogorov_smirnov(reference: xarray.DataArray, target: xarray.DataArray, to_return="pvalue")\
        -> Union[xarray.DataArray, xarray.Dataset]:
    """
    Description: Compute the KS statistic of the full domain and return the pvalue .

    Functions
    :param reference:
    :param target:
    :param to_return: select between returning the 'pvalue', the 'statistic' as a data Arrays or 'both' as a Dataset
    :return:
    """

    ks_statistic = reference.copy(deep=True)

    non_temporal_dimensions = [d for d in ks_statistic.dims if d != "time"]
    selector = {dim: reference[dim][0] for dim in non_temporal_dimensions}
    ks_statistic = reference.sel(selector)

    ks_pvalue = ks_statistic.copy(deep=True)

    if "time" in reference.dims:
        times = reference["time"]
        ks_results = [
            ks_2samp(target.sel(time=t).values.ravel(), reference.sel(time=t).values.ravel(), alternative='two-sided',
                     mode='auto') for t in times]
        statistic = [ks.statistic for ks in ks_results]
        p_value = [ks.pvalue for ks in ks_results]

    else:
        statistic, p_value = ks_2samp(target.ravel(), reference.ravel(), alternative='two-sided', mode='auto')

    # Put the results of the KS
    ks_pvalue.values = p_value
    ks_statistic.values = statistic

    # Fix attributes
    ks_statistic = fix_attributes(ks_statistic, "ks_statistic")
    ks_pvalue = fix_attributes(ks_pvalue, "ks_pvalue")

    # Combine data arrays in dataset
    result = xarray.Dataset(dict(statistic=ks_statistic, pvalue=ks_pvalue))
    if to_return in ["pvalue", "statistic"]:
        return result[to_return]
    elif to_return == "both":
        return result
    else:
        raise EnstoolsError(
            f"the parameter 'to_return' for kolmogorov_smirnov accepts the options 'pvalue', 'statistic' or 'both'. "
            f"{to_return!r} was provided.")


def kolmogorov_smirnov_index(reference: xarray.DataArray, target: xarray.DataArray, to_return="pvalue")\
        -> Union[xarray.DataArray, xarray.Dataset]:
    """
    Description: Compute the KS statistic of the full domain and return the -log10(1 - pvalue) .

    Functions
    :param reference:
    :param target:
    :param to_return: select between returning the 'pvalue', the 'statistic' as a data Arrays or 'both' as a Dataset
    :return:
    """
    ks = kolmogorov_smirnov(reference, target)
    return xarray.where(ks < 1.0, - numpy.log10(1 - ks), numpy.inf)



import logging
import warnings
import numpy as np
from math import sqrt
from enstools.plot.core import get_coordinates_from_xarray

from .backend import InteractiveBackend, Backend, Stream, Function
from .backend_plotly import InteractiveBackendPlotly
from .backend_bokeh import InteractiveBackendBokeh
from .backend3D import InteractiveBackend3D

import plotly.graph_objs as go
import plotly.figure_factory as ff 

from bokeh.plotting import figure
from bokeh.models import ColumnDataSource, Legend

from ...core import check_arguments


@check_arguments(dims={'variable': ('lat', 'lon')})
def interactive_contours(variable, lon=None, lat=None, **kwargs):
    """
    Creates a plot of interactive contours.

    Parameters
    ----------
    variable : xarray
            Values of a variable and a range of other parameters describing the variable and the grid.

    lon : xarray.DataArray or np.ndarray or str
            longitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.

    lat : xarray.DataArray or np.ndarray or str
            latitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.


    Other optional keyword arguments:

    **kwargs
            *figure*: bokeh or plotly figure object
                If provided, this figure instance will be used (and returned), otherwise a new
                figure will be created.

            *filled*: [*True* | *False*]
                If True a filled contour is plotted, which is the default

            *selected_backend* : Enum element, optional, default=Library.BOKEH
                Definition of a library for plotting. Element from Library Enum.
                 
            *aspect* : float, optional, default=None
                Aspect ratio between a figure width and a height as a fraction.
        
            *size* : int, optional, default=800
                Size of a figure width in pixels. Size of a figure height is calculated from an aspectratio.
        
            *map_resolution* : {'110m', '50m', '10m'}, optional, default='110m'
                Definition of a map resolution.

            *cmap* : str
                Name of the color map to use. Default: CMRmap_r

            *line_width* : int
                Line width, default is 1.
                      
    Returns
    -------
    figure
        Figure object contains interactive contours with a colorbars.            
    """

    # initialize a backend object based on the backend argument
    selected_backend = kwargs.pop("backend", Backend.BOKEH)
    if selected_backend == Backend.PLOTLY:
        _backend = InteractiveBackendPlotly()       
    elif selected_backend == Backend.BOKEH:
        _backend = InteractiveBackendBokeh()        
    else:
        raise ValueError(f"unsupported backend: {selected_backend}! backend must be an instance of Library Enum: PLOTLY or BOKEH.")
    
    # add warnings
    aspect_ratio = kwargs.pop("aspect", None)
    size = kwargs.pop("size", 800)
    previous_figure = kwargs.pop("figure", None)
    if previous_figure is not None:
        if aspect_ratio is not None or size != 800:
            warnings.warn("Change of figure size parameters is impossible for a new figure. Figure size parameters stay the same as in a previous figure.")

    map_resolution = kwargs.pop("map_resolution", '110m')
    if map_resolution not in {'110m', '50m', '10m'}:
        raise ValueError(f"unsupported map resolution: {map_resolution}! map_resolution must be an instance of: '110m', '50m' or '10m'.")
        
    # get the coordinates
    lon, lat = get_coordinates_from_xarray(variable, lon, lat, rad2deg=False)

    # interpolate to regular lat-lon grid if on icosahedral
    if variable.ndim == 1 and lon.ndim == 1 and lat.ndim == 1:
        variable = _backend.icosahedral_to_regular_coords(variable, clon=lon, clat=lat)
        lon, lat = get_coordinates_from_xarray(variable, create_mesh=False)

    # create a plot with filled contour lines?
    filled_contours = kwargs.pop('filled', True)

    # define figure parameters
    x_colorbar = None    # colorbar position on X-axis
    default_cmap = "CMRmap_r"
    if selected_backend == Backend.PLOTLY and filled_contours:
        default_cmap = "Inferno_r"
    colormap = kwargs.pop("cmap", default_cmap)
    if previous_figure is not None:
        fig = previous_figure
        x_colorbar = 1.2
        colormap = "Greys"

    variable_name = variable._name
        
    # define map borders
    lon0_map = float(lon.min())
    lon1_map = float(lon.max())
    lat0_map = float(lat.min())
    lat1_map = float(lat.max())

    # define an aspect ratio
    if aspect_ratio is None:
        aspect_ratio = _backend.get_aspect_ratio(lon0_map, lon1_map, lat0_map, lat1_map)

    # make paths for coastines and country borders
    xs_coas, ys_coas = _backend.make_path_map('coastlines', map_resolution)
    xs_bord, ys_bord = _backend.make_path_map('borders', map_resolution)

    # width for contour lines
    line_width = kwargs.pop("line_width", 1)

    # define contours and its colorscale for unfilled contours
    if not filled_contours:
        xs, ys, colors, values = _backend.get_contour_data(lon, lat, variable,
                                                           colormap, filled_contours=filled_contours)
        colorset, valueset, tick_interval = _backend.generate_colorscale(colors, values)

    if selected_backend == Backend.PLOTLY:
            
        # create a Plotly objects for coastlines and country borders
        traces_coas = _backend.make_scattergl_map(xs_coas, ys_coas)
        traces_bord = _backend.make_scattergl_map(xs_bord, ys_bord)
        base = traces_coas + traces_bord
            
        # define figure paramerets
        if previous_figure == None:
            fig = go.Figure(data=base)
            fig.update_layout(autosize=False, width=size, height=size / aspect_ratio,
                              xaxis_title="lon", yaxis_title="lat", plot_bgcolor="#FFFFFF",
                              xaxis=dict(range=[lon0_map, lon1_map], showgrid=False),
                              yaxis=dict(range=[lat0_map, lat1_map], showgrid=False),
                              legend=dict(orientation="h", bordercolor="#E0E0E0", borderwidth=1, 
                                          yanchor="top", y=1.15, xanchor="left", x=0.01))
            
        if filled_contours:
            
            # define and add filled contours to a figure
            contours = _backend.make_filled_contour(lon, lat, variable, variable_name, colormap, x_colorbar)
            fig.add_trace(contours)
            
        else:
            
            # add unfilled contours to a figure
            fig.add_trace(_backend.make_unfilled_contour(xs[0], ys[0], colors[0], variable_name, 
                                                         variable_name, showlegend=True))
            for i in range(1,len(xs)):
                fig.add_trace(_backend.make_unfilled_contour(xs[i], ys[i], colors[i], values[i], 
                                                             variable_name, showlegend=False))
            # define and add a colorbar to a figure    
            colorbar = _backend.make_colorbar(colorset, valueset, tick_interval, variable_name, x_colorbar)
            fig.add_trace(colorbar)

    elif selected_backend == Backend.BOKEH:
            
        # define figure parameters
        if previous_figure is None:
            fig = figure(aspect_ratio=aspect_ratio, plot_width=size,
                         x_range=(lon0_map, lon1_map), y_range=(lat0_map, lat1_map),
                         x_axis_label="lon", y_axis_label="lat", output_backend="webgl")
            fig.xgrid.visible = False
            fig.ygrid.visible = False
            fig.outline_line_color = None 
            
        if filled_contours:
            
            # define and add filled contours to a figure
            xs, ys, colors, values = _backend.get_contour_data(lon, lat, variable, colormap, filled_contours=filled_contours)
            source = ColumnDataSource(data={'xs': xs, 'ys': ys, 'line_color': colors})
            contours = fig.patches(xs='xs', ys='ys', alpha=1, line_width=0,
                                   fill_color='line_color', line_color='line_color', source=source)
            # define a colorbar
            colorset, valueset, tick_interval = _backend.generate_colorscale(colors, values)
            colorbar = _backend.make_colorbar(colorset, valueset, tick_interval, variable_name)

        else:
            
            # define and add unfilled contours to a figure
            source = ColumnDataSource(data={'xs': xs, 'ys': ys, 'line_color': colors})
            contours = fig.multi_line(xs='xs', ys='ys', line_color='line_color', line_width=line_width, source=source)
            
            # define a colorbar
            colorbar = _backend.make_colorbar(colorset, valueset, tick_interval, variable_name)

        # add coaslines and country borders to a figure
        fig.multi_line(xs_coas, ys_coas, line_color="grey")
        fig.multi_line(xs_bord, ys_bord, line_color="grey")
            
        # define figure layout parameters
        legend = Legend(items=[(variable_name , [contours])], location="top_left")
        fig.add_layout(colorbar, "right")
        fig.add_layout(legend, "above")
        fig.legend.click_policy = "hide"

    # check if all keyword arguments have been used.
    if len(kwargs) > 0:
        logging.warning("unknown arguments: %s" % ",".join(kwargs.keys()))
    return fig


@check_arguments(dims={'variable_u': ('lat', 'lon'), 'variable_v': ('lat', 'lon')})
def interactive_streamlines(variable_u, variable_v, lon=None, lat=None, **kwargs):
    """
    Creates a plot of interactive streamlines or vectors.

    Parameters
    ----------
    variable_u, variable_v : xarray.DataArray
        u- and v-component of streamline data.
        
    lon : xarray.DataArray or np.ndarray or str
            longitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.

    lat : xarray.DataArray or np.ndarray or str
            latitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.


    Other optional keyword arguments:

    **kwargs
            *figure*: bokeh or plotly figure object
                If provided, this figure instance will be used (and returned), otherwise a new
                figure will be created.
              
            *selected_backend* : Enum element, optional, default=Library.BOKEH
                Definition of a library for plotting. Element from Library Enum.

            *aspect* : float, optional, default=None
                Aspect ratio between a figure width and a height as a fraction.

            *size* : int, optional, default=800
                Size of a figure width in pixels. Size of a figure height is calculated from an aspectratio.

            *map_resolution* : {'110m', '50m', '10m'}, optional, default='110m'
                Definition of a map resolution.

            *line_type* : Enum element, optional, default=Stream.LINE
                Definition of a line type. Element from Stream Enum.
        
            *density* : float, optional, default=2
                Definition of a closeness of streamlines. When density = 1, the domain is divided into a 30x30 grid.

            *scale* : float, optional, default=0.02
                Possibility to scale a size of arrows(for example, to avoid overlap).
                      
    Returns
    -------
    figure
        Figure object contains interactive streamlines or vectors.              
    """
    
    # initialize a backend object based on the backend argument
    selected_backend = kwargs.pop("backend", Backend.BOKEH)
    if selected_backend == Backend.PLOTLY:
        _backend = InteractiveBackendPlotly()
    elif selected_backend == Backend.BOKEH:
        _backend = InteractiveBackendBokeh()
    else:
        raise ValueError(f"unsupported backend: {selected_backend}! backend must be an instance of Library Enum: PLOTLY or BOKEH.")
    
    # add warnings
    aspect_ratio = kwargs.pop("aspect", None)
    size = kwargs.pop("size", 800)
    previous_figure = kwargs.pop("figure", None)
    map_resolution = kwargs.pop("map_resolution", '110m')
    if previous_figure is not None:
        fig = previous_figure
        if aspect_ratio is not None or size != 800 or map_resolution != '110m':
            warnings.warn("Change of figure size parameters is impossible for a new figure. Figure size parameters stay the same as in a previous figure.")
            
    if map_resolution not in {'110m', '50m', '10m'}:
        raise ValueError(f"unsupported map resolution: {map_resolution}! map resolution must be an instance of: '110m', '50m' or '10m'.")
        
    # get the coordinates
    lon, lat = get_coordinates_from_xarray(variable_u, lon, lat, rad2deg=False)

    # interpolate to regular lat-lon grid if on icosahedral
    if variable_u.ndim == 1 and lon.ndim == 1 and lat.ndim == 1:
        variable_u = _backend.icosahedral_to_regular_coords(variable_u, clon=lon, clat=lat)
        variable_v = _backend.icosahedral_to_regular_coords(variable_v, clon=lon, clat=lat)
        lon, lat = get_coordinates_from_xarray(variable_u, create_mesh=False)

    # define figure parameters
    color = "black"

    # define map borders
    lon0_map = float(lon.min())
    lon1_map = float(lon.max())
    lat0_map = float(lat.min())
    lat1_map = float(lat.max())
    
    # define an aspectratio
    if aspect_ratio is None:
        aspect_ratio = _backend.get_aspect_ratio(lon0_map, lon1_map, lat0_map, lat1_map)
    
    # make paths for coastines and country borders
    xs_coas, ys_coas = _backend.make_path_map('coastlines', map_resolution)
    xs_bord, ys_bord = _backend.make_path_map('borders', map_resolution)
    
    # define streamlines paths
    line_type = kwargs.pop("line_type", Stream.LINE)
    density = kwargs.pop("density", 2)
    scale = kwargs.pop("scale", 0.02)
    if line_type == Stream.LINE:
        streamlines = _backend.get_streamline_data(lon, lat, variable_u, variable_v, density)
        
    # reduce a number of grid cells for vectors and get reference arrow data
    elif line_type == Stream.VECTOR:
        lons, lats, us, vs = _backend.reduce_vector_data(lon, lat, variable_u, variable_v, density)

        # increase range of Y-axis for placing of a reference arrow 
        lat1_map_arrow = abs(0.1*(lat1_map-lat0_map))
        lat1_map = lat1_map + lat1_map_arrow
        
        ref_arrow = _backend.get_reference_arrow_data(us, vs)
        if hasattr(variable_u, 'units'):
            ref_arrow_string = (str(ref_arrow) + ' ' + variable_u.units)
        else:
            ref_arrow_string = str(ref_arrow)
        
    else:
        raise ValueError(f"unsupported lines type: {line_type}! line type must be an instance of Stream Enum: LINE or VECTOR.")
            
    if selected_backend == Backend.PLOTLY:
        # create a Plotly objects for coastlines and contry borders
        traces_coas = _backend.make_scattergl_map(xs_coas, ys_coas)
        traces_bord = _backend.make_scattergl_map(xs_bord, ys_bord)
        base = traces_coas + traces_bord
            
        # define figure parameters
        if previous_figure is None:
            fig = go.Figure(data=base)
            fig.update_layout(autosize=False, width=size, height=size/aspect_ratio,
                              xaxis_title="lon", yaxis_title="lat", plot_bgcolor="#FFFFFF", 
                              xaxis=dict(range=[lon0_map, lon1_map], showgrid=False), 
                              yaxis=dict(range=[lat0_map, lat1_map], showgrid=False), 
                              legend=dict(orientation="h", bordercolor='#E0E0E0', borderwidth=1, 
                                          yanchor="top", y=1.20, xanchor="left", x=0.01))
            
        if line_type == Stream.LINE:
            # add streamlines to a figure   
            fig.add_trace(_backend.make_scattergl_streamline(streamlines[0], color, showlegend=True))
            for i in range(len(streamlines)):
                fig.add_trace(_backend.make_scattergl_streamline(streamlines[i], color, showlegend=False))
                
        elif line_type == Stream.VECTOR:
            # define a reference arrow
            fig_ref_arrow_temp = (ff.create_quiver([lon0_map + 0.5], [lat1_map - 0.8*lat1_map_arrow], [ref_arrow], [0], 
                                                   scale=scale, arrow_scale=0.3, line_color=color, 
                                                   name="reference arrow", legendgroup="vectors", showlegend=False))
            
            # define vectors
            fig_temp = (ff.create_quiver(lons, lats, us, vs, scale=scale, arrow_scale=0.3, line_color=color,
                                         name="vectors", legendgroup="vectors"))
            
            # add vectors and a reference arrow to a figure
            fig.add_traces(data = fig_temp.data)
            fig.add_traces(data = fig_ref_arrow_temp.data)
            fig.add_trace(go.Scatter(x=[lon0_map + 0.5], y=[lat1_map - 0.8*lat1_map_arrow],
                                     mode="text", text=ref_arrow_string, textposition="top right",
                                     name="reference arrow", legendgroup="vectors", showlegend=False))
            
    elif selected_backend == Backend.BOKEH:
        # define figure parameters
        if previous_figure is None:
            fig = figure(aspect_ratio=aspect_ratio, plot_width=size,
                         x_range=(lon0_map, lon1_map), y_range=(lat0_map, lat1_map), 
                         x_axis_label="lon", y_axis_label="lat", output_backend="webgl")
            fig.xgrid.visible = False
            fig.ygrid.visible = False
            fig.outline_line_color = None      
        
        if line_type == Stream.LINE:
            # separate and add streamlines to a figure
            xs, ys = _backend.separate_streamline_data(streamlines)
            
            lines = fig.multi_line(xs, ys, color=color, line_width=2, line_alpha=0.8) 
            
            legend = Legend(items=[("streamlines" , [lines])], location="top_left")
        
        elif line_type == Stream.VECTOR:
            # define and add vectors and a reference arrow to a figure
            xs, ys, xs_ref_arrow, ys_ref_arrow = _backend.separate_vector_data(lons, lats, us, vs, scale, 
                                                                               ref_arrow, lon0_map, lat1_map, lat1_map_arrow)  
            
            lines = fig.multi_line(xs, ys, color=color, line_width=2)
            
            ref_arrow_text = fig.text(x=[lon0_map+0.5], y=[lat1_map-0.8*lat1_map_arrow], text=[ref_arrow_string])
            ref_arrow = fig.multi_line(xs_ref_arrow, ys_ref_arrow, color=color, line_width=2)
                                                      
            legend = Legend(items=[("vectors" , [lines, ref_arrow_text, ref_arrow])], location="top_left")     
            
        # add coastlines and country borders to a figure
        fig.multi_line(xs_coas, ys_coas, line_color="grey")
        fig.multi_line(xs_bord, ys_bord, line_color="grey")
        
        # define figure layout parameters
        fig.add_layout(legend, "above")
        fig.legend.click_policy = "hide"
        
    # check if all keyword arguments have been used.
    if len(kwargs) > 0:
        logging.warning("unknown arguments: %s" % ",".join(kwargs.keys()))
    return fig


def __individual_grid_cells(data, grid, *kwargs, regional = True, without_edges = True,
                          with_map = True,  map_resolution='110m', function = Function.TRISURF, 
                          aspectratio = 1., size = 800):
    """
    Creates an interactive 3D plot of individual grid cells.

    Parameters
    ----------
    data : xarray
        Values of a variable and a range of other parameters describing a variable and a grid.
            
    grid : xarray
        Parameters of grid cells.
        
    *kwargs
        Possibility to define a colormap.
                                          
    regional : bool, optional, default=True
        Definition of a scale. Argument equals to True for a regional scale and to False for a global scale(=whole Earth).
        
    without_edges : bool, optional, default=True
        Definition of a type of borders between individual grid cells. Argument equals to True for invisible edges between cells.
        
    with_map : bool, optional, default=True
        Possibility to add map contours above a plot.
        
    map_resolution : {'110m', '50m', '10m'}, optional, default='110m'
        Definition of a map resolution.
                 
    function : Enum element, optional, default=Function.TRISURF
        Definition of a Plotly function used for plotting. Element from Function Enum.
        
    aspectratio : float, optional, default=1.
        Ratio between a figure width and a height as a fraction. 
        
    size : int, optional, default=800
        Size of a figure width in pixels. Size of a figure height is calculated from an aspectratio.
                      
    Returns
    -------
    figure 
        Figure object contains an interactive 3D plot of individual grid cells.             
    """    
    
    # initialize a backend object
    _backend = InteractiveBackend3D()
    
    # check possibility to use given function arguments(for cells edges)
    if not without_edges:
        plot_edges = True
        if not regional:
            warnings.warn("Visualization of edges of grid cells is impossible for global scale! Cells edges are invisible at the plot.")
            without_edges = True
            plot_edges = False
        elif function == Function.MESH3D:
            warnings.warn("Visualization of edges of grid cells is impossible for MESH3D! Default function TRISURF was used instead.")
            function = Function.TRISURF
    else:
        plot_edges = False  
        
    if map_resolution not in {'110m', '50m', '10m'}:
        raise ValueError(f"unsupported map resilution: {map_resolution}! map resolution must be an instance of: '110m', '50m' or '10m'.")
        
    # define individual grid cells X- and Y-coordinates
    lons = grid.longitude_vertices.values
    lats = grid.latitude_vertices.values
    
    # define region borders
    lon0_map = np.min(np.degrees(lons)); lon1_map = np.max(np.degrees(lons))
    lat0_map = np.min(np.degrees(lats)); lat1_map = np.max(np.degrees(lats))
    
    # define approximate cell size 
    cell_size = sqrt(4*np.max(grid.cell_area_p.values)/sqrt(3))
    
    # define map paths 
    if with_map:
        xs_coas, ys_coas, zs_coas, n_coas = _backend.make_path_map3D('coastlines', cell_size, lons, lats, regional,
                                                                     lon0_map, lon1_map, lat0_map, lat1_map, map_resolution)
        xs_bord, ys_bord, zs_bord, n_bord = _backend.make_path_map3D('borders', cell_size, lons, lats, regional, 
                                                                     lon0_map, lon1_map, lat0_map, lat1_map, map_resolution) 
        
    # transform grid cells coordinates to cartesian coordinates 
    cartesian_coordinates = _backend.spherical_to_cartesian(lons, lats, radius=6371229.0)    
    xs = cartesian_coordinates.T[0]
    ys = cartesian_coordinates.T[1]
    zs = cartesian_coordinates.T[2]

    xratio, yratio, zratio = _backend.set_aspectratio3D(xs, ys, zs)
    
    # define a colormap
    colormap = 'jet'
    if len(kwargs) != 0:
        colormap = kwargs[0]
    
    # define colors of individual grid cells
    colors = _backend.set_colors(data[0,0,:].values, colormap, function) 
    
    # define figure parameters
    # define a colorbar
    colorbar = _backend.make_colorbar_continuous(data[0,0,:].values, colormap)
    
    # define layout parameters
    layout_base = {'xaxis': { 'showgrid': False, 'zeroline': False, 'visible': False},
                   'yaxis': { 'showgrid': False, 'zeroline': False, 'visible': False}}  
    scene = {'xaxis': { 'showgrid': False, 'zeroline': False, 'visible': False},
             'yaxis': { 'showgrid': False, 'zeroline': False, 'visible': False},
             'zaxis': { 'showgrid': False, 'zeroline': False, 'visible': False}}
    
    # define a legend position 
    legend = dict(yanchor="top", xanchor="right")
    
    if function == Function.TRISURF:
        
        # define individual grid cells and add them to a figure 
        simplices = grid.vertex_of_cell.T.values - 1
        fig = go.Figure(data=ff.create_trisurf(x=xs, y=ys, z=zs, color_func=colors, simplices=simplices, 
                                               plot_edges=plot_edges, aspectratio=dict(x=xratio, y=yratio, z=zratio)))
        
    elif function == Function.MESH3D:
        
        # define individual grid cells and add them to a figure 
        i = grid.vertex_of_cell.values[0] - 1
        j = grid.vertex_of_cell.values[1] - 1
        k = grid.vertex_of_cell.values[2] - 1
        fig = go.Figure(data = go.Mesh3d(x=xs, y=ys, z=zs, facecolor=colors, i=i, j=j, k=k))
        
    else:
        raise ValueError(f"unsupported function: {function}! function must be an instance of Function Enum: TRISURF or MESH3D.")
        
    if with_map:
  
        # add coastlines to a figure
        fig.add_trace(go.Scatter3d(x=xs_coas[0], y=ys_coas[0], z=zs_coas[0], mode='lines', 
                                   legendgroup='map', name='map', showlegend=True, line=dict(color='grey', width=2)))
        for i in range(1, n_coas): 
            fig.add_trace(go.Scatter3d(x=xs_coas[i], y=ys_coas[i], z=zs_coas[i], mode='lines', 
                                       legendgroup='map', name='map', showlegend=False, line=dict(color='grey', width=2)))
            
        # add country borders to a figure
        for i in range(n_bord): 
            fig.add_trace(go.Scatter3d(x=xs_bord[i], y=ys_bord[i], z=zs_bord[i], mode='lines', 
                                       legendgroup='map', name='map', showlegend=False, line=dict(color='grey', width=2)))
        
    # add a colorbar 
    fig.add_trace(colorbar)
        
    # update a layout
    fig.update_layout(layout_base)
    fig.update_layout(width=size, height=size/aspectratio, scene=scene, legend=legend, plot_bgcolor="#FFFFFF")
    
    # define a camera eye and a camera position for regional scale
    if regional:
        camera_eye = _backend.set_camera_eye(lons, lats)
        camera = dict(up=dict(x=0, y=0, z=1), center=dict(x=0, y=0, z=0), 
                      eye=dict(x=1.7*camera_eye[0], y=1.7*camera_eye[1], z=camera_eye[2]))
        fig.update_layout(scene_camera = camera, scene_dragmode='pan') 
    
    return fig


def __interactive_3Dcontours(data, previous_figure, *kwargs, grid = None):
    """
    Creates an interactive 3D contours on already existing plot of individual grid cells.

    Parameters
    ----------
    data : xarray
        Values of a variable and a range of other parameters describing a variable and a grid.
            
    previous_figure : figure
        Pre-created figure of individual grid cells.
        
    *kwargs
        Possibility to define a colormap of contours.
        
    grid : xarray, optional, default=None
        Parameters of grid cells for icosahedral coordinates.
                      
    Returns
    -------
    figure 
        Figure object contains an interactive 3D plot of individual grid cells with contours.             
    """    
    
    # initialize a backend object
    _backend = InteractiveBackend3D()
    
    # if coordinates are icosahedral: convert to regular-lat-lon coordinates
    if grid != None:
        data = _backend.icosahedral_to_regular_coords(grid, data)
        
    # define names of regular-lat.lon coordinates
    lon, lat = _backend.coords_name(data)
    
    # define a base figure 
    fig = previous_figure
    
    # define contours parameters
    variable_name = data._name    
    x_colorbar = 1.2
    colormap = "BuPu"
    if len(kwargs) != 0:
        colormap = kwargs[0]
    
    # define contours and its colorbar
    xs_degrees, ys_degrees, colors, values = _backend.get_contour_data(data[lon].values, data[lat].values, data[0,0,:,:].values,
                                                                       colormap, filled_contours=False)
    
    xs, ys, zs, npaths = _backend.paths_cartesian(xs_degrees, ys_degrees, radius=6401229.0)
    
    colorset, valueset, tick_interval = _backend.generate_colorscale(colors, values)
    
    colorbar = _backend.make_colorbar(colorset, valueset, tick_interval, variable_name, x_colorbar)
    
    # add contours and a colorbar to a figure
    fig.add_trace(go.Scatter3d(x=xs[0], y=ys[0], z=zs[0], mode='lines', legendgroup=variable_name, name=variable_name,
                               showlegend=True, line=dict(color=colors[0],width=4)))
    
    for i in range(1, npaths): 
        fig.add_trace(go.Scatter3d(x=xs[i], y=ys[i], z=zs[i], mode='lines', legendgroup=variable_name, name=variable_name,
                                   showlegend=False, line=dict(color=colors[i],width=4)))
        
    fig.add_trace(colorbar) 
    
    return fig


def __interactive_3Dstreamlines(data_horizontal, data_vertical, previous_figure, *kwargs, linetype = Stream.LINE, grid = None,
                              density = 2, scale = 0.02):
    """
    Creates an interactive 3D streamlines or vectors on already existing plot of individual grid cells.

    Parameters
    ----------
    datau, datav : xarray
        Values of a variable and a range of other parameters describing a variable and a grid.
        
    linetype : Enum element, optional, default=Stream.LINE
        Definition of a linetype. Element from Stream Enum.
            
    previous_figure : figure
        Pre-created figure of individual grid cells.
        
    *kwargs
        Possibility to define a color of streamlines or vectors.
        
    grid : xarray, optional, default=None
        Parameters of grid cells.
        
    density : float, optional, default=2
        Controls the closeness of streamlines. When density = 1, the domain is divided into a 30x30 grid.
                     
    scale : float, optional, default=0.02
        Possibility to scale a size of arrows(ideally to avoid overlap).
                      
    Returns
    -------
    figure 
        Figure object contains an interactive 3D plot of individual grid cells with streamlines or vectors.             
    """   
    
    # initialize a backend object
    _backend = InteractiveBackend3D()
    _backend_bokeh = InteractiveBackendBokeh()  

    # if coordinates are icosahedral: convert to regular-lat-lon coordinates
    if grid != None:
        data_horizontal = _backend.icosahedral_to_regular_coords(grid, data_horizontal)
        data_vertical   = _backend.icosahedral_to_regular_coords(grid, data_vertical)
        
    # define names of regular-lat.lon coordinates
    lon, lat = _backend.coords_name(data_horizontal)
        
    # define lines borders
    lon0 = float(np.min(data_horizontal[lon])); lon1 = float(np.max(data_horizontal[lon]))
    lat0 = float(np.min(data_horizontal[lat])); lat1 = float(np.max(data_horizontal[lat]))
    
    # define a base figure 
    fig = previous_figure
    
    # define figure parameters
    color = "black"
    if len(kwargs) != 0:
        color = kwargs[0]
    
    # define streamlines paths
    if linetype == Stream.LINE:
        
        line_name = 'streamlines'
        
        streamlines = _backend.get_streamline_data(data_horizontal[lon].values, data_horizontal[lat].values, 
                                                   data_horizontal[0,0,:,:].values, data_vertical[0,0,:,:].values, density)
        
        xs_degrees, ys_degrees = _backend_bokeh.separate_streamline_data(streamlines)
        
        # transform to cartesian coordinates and separate coordinates
        xs, ys, zs, npaths = _backend.paths_cartesian(xs_degrees, ys_degrees, radius=6401229.0)
        
    # reduce a number of grid cells for vectors and define a reference arrow
    elif linetype == Stream.VECTOR:
        
        line_name = 'vectors'
        
        # reduce a number of grid cells
        lons, lats, us, vs = _backend.reduce_vector_data(data_horizontal[lon].values, data_horizontal[lat].values,
                                                         data_horizontal[0,0,:,:].values, data_vertical[0,0,:,:].values, density)
        
        # define a position of a reference arrow and its length
        lat1_arrow = abs(int(0.1*(lat1-lat0)))
        lat1 = lat1 + lat1_arrow
        ref_arrow = _backend.get_reference_arrow_data(us, vs)
        ref_arrow_string= (str(ref_arrow) + ' ' + data_horizontal.units)
        
        # define vectors and a reference arrow
        xs_degrees, ys_degrees, xs_ref_arrow_degrees, ys_ref_arrow_degrees = _backend_bokeh.separate_vector_data(lons, lats, us, vs,
                                                                                                                 scale, ref_arrow,
                                                                                                                 lon0, lat1, 0)
        # transform to cartesian coordinates and separate coordinates
        xs, ys, zs, npaths = _backend.paths_cartesian(xs_degrees, ys_degrees, radius=6401229.0)
        xs_ref_arrow, ys_ref_arrow, zs_ref_arrow, npaths_ref_arrow = _backend.paths_cartesian(xs_ref_arrow_degrees,
                                                                                              ys_ref_arrow_degrees,
                                                                                              radius=(1+0.01*lat1_arrow)*6401229.0)       
    else:
        raise ValueError(f"unsupported lines type: {linetype}! line type must be an instance of Stream Enum: LINE or VECTOR.")
    
    # add streamlines to a figure   
    fig.add_trace(go.Scatter3d(x=xs[0], y=ys[0], z=zs[0], mode='lines', legendgroup=line_name, name=line_name,
                               showlegend=True, line=dict(color=color,width=4)))
    for i in range(1, npaths): 
        fig.add_trace(go.Scatter3d(x=xs[i], y=ys[i], z=zs[i], mode='lines', legendgroup=line_name, name=line_name,
                                   showlegend=False, line=dict(color=color,width=4))) 
    
    # add reference arrow for vectors
    if linetype == Stream.VECTOR:
        
        for i in range(npaths_ref_arrow):
            fig.add_trace(go.Scatter3d(x=xs_ref_arrow[i], y=ys_ref_arrow[i], z=zs_ref_arrow[i], mode='lines', legendgroup=line_name,
                                       name=line_name, showlegend=False, line=dict(color=color,width=4))) 
        
        fig.update_layout(scene=dict(annotations=[dict(showarrow=False, 
                                                       x=xs_ref_arrow[0][0], y=ys_ref_arrow[0][0], 
                                                       z=(1 + 0.005*lat1_arrow)*zs_ref_arrow[0][0],
                                                       text=ref_arrow_string)]))
    
    return fig




class EnstoolsError(Exception):
    """Base class for Enstools Errors"""


class WrongCompressionSpecificationError(EnstoolsError):
    """Raised when the compression specification is not valid"""


class WrongCompressionModeError(WrongCompressionSpecificationError):
    """Raised when the compression mode is not valid"""



import numpy as np
import xarray
from skimage.metrics import structural_similarity

from enstools.core.errors import EnstoolsError

horizontal_spatial_dimensions = ["lat", "lon"]


def check_coordinates(data_array: xarray.DataArray):
    """
    Check that the provided data has the horizontal spatial coordinates.
    """
    missing_dimensions = [d for d in horizontal_spatial_dimensions if d not in data_array.dims]
    if missing_dimensions:
        message = f"Structural similarity index expects data with dimensions {horizontal_spatial_dimensions}.\n" \
                  f"Dimension/s {missing_dimensions} missing."
        raise EnstoolsError(message)


def structural_similarity_index(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    """
    Description: A meaningful description of the metric with references if necessary would be nice.

    Functions
    :param reference:
    :param target:
    :return:
    """
    """
    Returns the SSIM of the full DataArray. The computation its done slice by slice.
    """
    check_coordinates(reference)

    # Load data
    reference.load()
    target.load()

    # Apply wrapper to horizontal slices
    res = xarray.apply_ufunc(compute_ssim_slice, reference, target,
                             #
                             vectorize=True,
                             input_core_dims=[horizontal_spatial_dimensions, horizontal_spatial_dimensions],
                             output_core_dims=[[]],
                             )
    pending_dimensions = [d for d in res.dims if d != "time"]

    if pending_dimensions:
        res = res.mean(dim=pending_dimensions)
    return res


def compute_ssim_slice(target: np.ndarray, reference: np.ndarray) -> float:
    """
    Returns the SSIM of a data slice. It uses the structural_similarity function from skimage.metrics.
    """
    ref_min, ref_max = np.min(reference), np.max(reference)
    target_min, target_max = np.min(target), np.max(target)

    true_min = min(ref_min, target_min)
    true_max = min(ref_max, target_max)

    return structural_similarity(target, reference, data_range=true_max - true_min)


def structural_similarity_log_index(reference: xarray.DataArray, target: xarray.DataArray) -> xarray.DataArray:
    ssim = structural_similarity_index(reference, target)

    return xarray.where(ssim >= 1.0, np.inf, -np.log10(1 - ssim))



import logging
from collections import OrderedDict
import xarray
import dask
import distributed
import six
import os
import re
import numpy as np
import glob
from typing import Union, Tuple, List
from pathlib import Path
from .paths import clean_paths
from enstools.misc import add_ensemble_dim, is_additional_coordinate_variable, first_element, \
    set_ensemble_member
from enstools.core import get_client_and_worker
from packaging import version

try:
    from enstools.encoding.api import check_dataset_filters_availability, check_filters_availability

    enstools_encoding_available = True
except ModuleNotFoundError:
    enstools_encoding_available = False
from .dataset import drop_unused
from .file_type import get_file_type

try:
    from .eccodes import read_grib_file
except ImportError:
    pass


def __read_one_file(filename: Path, constant=None, decode_times=True, **kwargs):
    """
    Read one or more input files

    Parameters
    ----------
    filename : Path object
            name of one individual file of unix shell-like file name pattern for multiple files

    constant : str
            name of a file containing constant variables.

    **kwargs
            all arguments accepted by xarray.open_dataset() of xarray.open_mfdataset()

    Returns
    -------
    xarray.Dataset
            in-memory representation of the content of the input file(s)
    """
    if constant is not None:
        return read(filename, constant=constant, **kwargs)
    else:
        # are we inside of a worker and is a distributed client available?
        client, worker = get_client_and_worker()
        # do we have a client, but we are not running inside of a worker?
        if client is not None and worker is None:
            return dask.compute(dask.delayed(read)(filename, **kwargs))[0]
        return __open_dataset(filename, client, worker, decode_times=decode_times, **kwargs)


def read(
        filenames: Union[List[str], Tuple[str], str, Path, List[Path]],
        constant=None,
        merge_same_size_dim=False,
        members_by_folder=False,
        member_by_filename=None,
        decode_times=True,
        **kwargs,
):
    """
    Read multiple input files

    Parameters
    ----------
    filenames : list of str or tuple of str or str
            names of individual files or filename pattern

    merge_same_size_dim : bool
            dimensions of the same size but with different names are merged together. This is sometimes useful to
            merge datasets from different file formats (e.g., grib and netcdf). Default: False.

    members_by_folder : bool
            interpret files in one subdirectory as data from the same ensemble member. This is useful if the ensemble
            member number if not stored within the input file. Defaut: False.

    member_by_filename : str
            string containing a regular expression to read to ensemble member from the file names. The first group
            within this regex has to match an integer. Example: member_by_filename="_P(\\\d+).nc" would be able the read the
            12 from the file name AROME-EPS_2012110500_P12.nc.

    constant : str
            name of a file containing constant variables.
    decode_times: bool
            decode the times
    **kwargs
            all arguments accepted by xarray.open_dataset() of xarray.open_mfdataset() plus some additional:

            *drop_unused*: bool
                remove unused coordinates. This improves the performance of the merge process. The merge process compares
                the coordinates from all files with each other.

            *in_memory*: bool
                store the complete arrays in memory. Data is still handled as dask arrays, but not backed by the input
                files anymore. This works of course only for datasets which fit into memory.

            *leadtime_from_filename*: bool
                COSMO-GRIB1-Files do not contain exact times. If this argument is set, then the timestamp is calculated
                from the init time and the lead time from the file name.

    Returns
    -------
    xarray.Dataset
            in-memory representation of the content of the input file(s)
    """
    if enstools_encoding_available:
        # we need to make sure that we are able to read compressed files
        if not check_filters_availability():
            import hdf5plugin

    filenames = clean_paths(filenames)

    # Hint: the \\\ the the docstring for member_by_filename is only included because the html-documentation is
    # otherwise not rendered correctly.
    # create at first a list of all input files
    datasets = []
    # loop over all input files and create delayed read objects
    expanded_filenames = []
    parent_folders = []
    # do we want to create an ensemble dimension?
    if member_by_filename is not None or members_by_folder is True:
        kwargs["create_ens_dim"] = True
    for filename in filenames:
        # is the filename a pattern?
        one_file = os.path.abspath(filename)
        datasets.append(dask.delayed(__read_one_file)(filename, decode_times=decode_times, **kwargs))
        expanded_filenames.append(filename)
        parent = os.path.dirname(filename)
        if not parent in parent_folders:
            parent_folders.append(parent)

    datasets = dask.compute(*datasets, traverse=False)

    # are there ensemble members in different folders?
    if members_by_folder and len(parent_folders) > 1:
        # sort all parent folders and assign ensemble member numbers to them
        parent_folders = sorted(parent_folders)
        ens_member_by_folder = __assign_ensemble_member_number_to_folders(parent_folders)

        # create the ensemble dimension within the datasets
        n_files_per_folder = {}
        for ids in range(len(datasets)):
            folder = os.path.dirname(expanded_filenames[ids])

            # count the files per member and remove incomplete members
            if folder in n_files_per_folder:
                n_files_per_folder[folder] += 1
            else:
                n_files_per_folder[folder] = 1

            # create missing ensemble dimension
            set_ensemble_member(datasets[ids], ens_member_by_folder[folder])

        # check the number of files per folder
        incomplete_folders = []
        max_files = 0
        for folder, n_files in six.iteritems(n_files_per_folder):
            if n_files > max_files:
                max_files = n_files
        for folder, n_files in six.iteritems(n_files_per_folder):
            if n_files < max_files:
                incomplete_folders.append(folder)

        # remove incomplete ensemble members
        if len(incomplete_folders) > 0:
            incomplete_folders.sort()
            for folder in incomplete_folders:
                logging.warning(
                    "The ensemble member in folder '%s' seems to be incomplete with only %d of %d files.This member will not be part of the merged dataset.",
                    folder, n_files_per_folder[folder], max_files)
            new_datasets = []
            for ids in range(len(datasets)):
                if os.path.dirname(expanded_filenames[ids]) not in incomplete_folders:
                    new_datasets.append(datasets[ids])
            datasets = new_datasets

    # create the ensemble dimension based on integers included in the filename
    if member_by_filename is not None:
        if members_by_folder:
            raise IOError("member_by_filename and members_by_folder are not allowed at the same time!")

        # create the ensemble dimension for all datasets
        n_files_per_member = {}
        for ids in range(len(datasets)):
            # find the member information within the filename
            match = re.search(member_by_filename, os.path.basename(expanded_filenames[ids]))
            if match is None:
                raise IOError("pattern for ensemble member ('%s') information not found in filename '%s'" % (
                    member_by_filename, os.path.basename(expanded_filenames[ids])))
            member = int(match.group(1))

            # create missing ensemble dimension
            set_ensemble_member(datasets[ids], member)

            # count the files per member and remove incomplete members
            if member in n_files_per_member:
                n_files_per_member[member] += 1
            else:
                n_files_per_member[member] = 1

            # check the number of files per folder
            incomplete_members = []
            max_files = 0
            for member, n_files in six.iteritems(n_files_per_member):
                if n_files > max_files:
                    max_files = n_files
            for member, n_files in six.iteritems(n_files_per_member):
                if n_files < max_files:
                    incomplete_members.append(member)

        # remove incomplete ensemble members
        if len(incomplete_members) > 0:
            incomplete_members.sort()
            for member in incomplete_members:
                logging.warning(
                    "The ensemble member '%s' seems to be incomplete with only %d of %d files. This member will not be part of the merged dataset.",
                    member, n_files_per_member[member], max_files)
            new_datasets = []
            for ids in range(len(datasets)):
                member = int(re.search(member_by_filename, os.path.basename(expanded_filenames[ids])).group(1))
                if member not in incomplete_members:
                    new_datasets.append(datasets[ids])
            datasets = new_datasets

    # if dimensions have the same size but different names, then merge them by renaming
    if merge_same_size_dim:
        datasets = __rename_same_size_dim(datasets)

    # combine datsets from different files
    result = __merge_datasets(datasets)

    # is there a file with constant data?
    if constant is not None:
        # read constant data
        constant_data = __open_dataset(constant, None, None, decode_times=decode_times)
        # remove time axes if present
        if "time" in constant_data.dims:
            constant_data = constant_data.isel(time=0)
        # remove variables also present in the non constant data
        for one_var in constant_data.data_vars:
            if one_var in result or one_var.lower() in result or one_var.upper() in result:
                logging.warning("variable '%s' in constant and non-constant data => renamed to '%s_constant'!", one_var,
                                one_var)
                constant_data = constant_data.rename({one_var: "%s_constant" % one_var})
        # rename dimensions?
        if merge_same_size_dim:
            datasets = __rename_same_size_dim([result, constant_data])
        else:
            datasets = [result, constant_data]
        # merge constant data
        if hasattr(xarray, "combine_by_coords"):
            result = xarray.combine_by_coords(datasets)
        else:
            result = xarray.auto_combine(datasets)

    if enstools_encoding_available:
        assert check_dataset_filters_availability(result), "The dataset uses some filters that are not available."

    return result


def __rename_same_size_dim(datasets):
    """
    Look for dimensions with the same size and rename them to the first name found.

    Parameters
    ----------
    datasets : list
            list of datasets to check

    Returns
    -------
    list:
            list of modified datasets
    """
    size_name_mapping = {}
    rename_mapping = {}
    for ds in datasets:
        for dim_name, dim_size in six.iteritems(ds.sizes):
            if dim_size not in size_name_mapping:
                size_name_mapping[dim_size] = dim_name
            else:
                if dim_name != size_name_mapping[dim_size]:
                    rename_mapping[dim_name] = size_name_mapping[dim_size]
    if len(rename_mapping) > 0:
        # we need do rename something. return a new list of datasets
        result = []
        for ds in datasets:
            rename_mapping_for_ds = {}
            for old_name, new_name in six.iteritems(rename_mapping):
                if old_name in ds.dims:
                    rename_mapping_for_ds[old_name] = new_name
            if len(rename_mapping_for_ds) > 0:
                result.append(ds.rename(rename_mapping_for_ds))
            else:
                result.append(ds)
        return result
    else:
        # nothing to be done. return the unchanged input
        return datasets


def __merge_datasets(datasets):
    """

    Parameters
    ----------
    datasets

    Returns
    -------

    """
    # anything to do?
    if len(datasets) == 1:
        return datasets[0]

    # variables not available in all files are merged separately
    datasets_incomplete = []
    vars_not_in_all_files = set()
    all_vars = set()
    for ds in datasets:
        for one_var in ds.data_vars.keys():
            if one_var not in all_vars:
                all_vars.add(one_var)
    for one_var in all_vars:
        for ds in datasets:
            if one_var not in ds.data_vars:
                vars_not_in_all_files.add(one_var)
    for one_var in vars_not_in_all_files:
        new_unmerged_ds = []
        for ids, ds in enumerate(datasets):
            if one_var in ds.data_vars:
                if len(ds.data_vars) == 1:
                    new_unmerged_ds.append(ds)
                else:
                    new_ds = xarray.Dataset()
                    new_ds[one_var] = ds[one_var]
                    del ds[one_var]
                    new_unmerged_ds.append(new_ds)
        if len(new_unmerged_ds) > 0:
            datasets_incomplete.append(__merge_datasets(new_unmerged_ds))

    # find common coordinates and sort variables according to those coordinates
    # get all coordinates
    all_coords = set()
    for ds in datasets:
        for coord in ds.coords.keys():
            if coord not in all_coords and (
                    np.issubdtype(ds.coords[coord].dtype, np.number) or np.issubdtype(ds.coords[coord].dtype,
                                                                                      np.datetime64)):
                all_coords.add(coord)
    # filter for coordinates present in all files
    coords_in_all_files = OrderedDict()
    for coord in all_coords:
        in_all = True
        ascending = True
        order_checked = False
        for ds in datasets:
            if coord in ds.coords:
                # multi-dim coords are not sortable
                if ds.coords[coord].ndim > 1:
                    in_all = False
                    break
                # the order is only checked once
                if not order_checked and ds.coords[coord].size > 1:
                    if ds.coords[coord][0] > ds.coords[coord][1]:
                        ascending = False
                    order_checked = True
            else:
                in_all = False
                break
        if in_all:
            coords_in_all_files[coord] = ascending

    # sort the dataset by all shared coordinates
    for coord, ascending in six.iteritems(coords_in_all_files):
        datasets = sorted(datasets, key=lambda x: first_element(x.coords[coord]), reverse=not ascending)

    # find coords with different values in different files
    for coord in six.iterkeys(coords_in_all_files):
        # loop over all files
        datasets_with_one_coord_value = {}
        for one_ds in datasets:
            first_value = first_element(one_ds.coords[coord])
            if first_value not in datasets_with_one_coord_value:
                datasets_with_one_coord_value[first_value] = [one_ds]
            else:
                datasets_with_one_coord_value[first_value].append(one_ds)
        # are there values with more than one dataset?
        different_values = sorted(datasets_with_one_coord_value.keys())
        if not coords_in_all_files[coord]:
            different_values = list(reversed(different_values))
        # merge files with more than one value
        if len(different_values) > 1:
            new_datasets = []
            for one_value in different_values:
                new_datasets.append(__merge_datasets(datasets_with_one_coord_value[one_value]))
            datasets = new_datasets

    # are there datasets with ensemble_members attribute?
    ensemble_members_found = []
    for one_dataset in datasets:
        if "ensemble_member" in one_dataset.attrs:
            if not one_dataset.attrs["ensemble_member"] in ensemble_members_found:
                ensemble_members_found.append(one_dataset.attrs["ensemble_member"])

    # data from more than one ensemble member was found. create the ensemble dimension for all datasets
    if len(ensemble_members_found) > 1:
        for one_dataset in datasets:
            if "ensemble_member" in one_dataset.attrs:
                add_ensemble_dim(one_dataset, one_dataset.attrs["ensemble_member"], inplace=True)
            else:
                logging.warning("Trying to merge ensemble and non-ensemble data. That will possibly fail!")
        result = __merge_datasets(datasets)
    else:
        # try to merge the datasets
        try:
            if hasattr(xarray, "combine_by_coords"):
                # the possibility to drop conflicting attributes was added in version 0.17.0
                if version.parse(xarray.__version__) >= version.parse('0.17.0'):
                    combine_attrs_methode = 'drop_conflicts'
                else:
                    combine_attrs_methode = 'override'
                result = xarray.combine_by_coords(list(datasets), combine_attrs=combine_attrs_methode)
            else:
                result = xarray.auto_combine(list(datasets))
        except Exception as ex:
            logging.error("merging the following datasets automatically failed:")
            for one_ds in datasets:
                logging.error("%s", one_ds)
                logging.error("-" * 80)
            raise ex

    # are the variables not available in all files?
    if len(datasets_incomplete) > 0:
        datasets_incomplete = xarray.merge(datasets_incomplete)
        result = result.merge(datasets_incomplete)

    return result


def __open_dataset(filename, client, worker, decode_times=True, **kwargs):
    """
    read one input file. the type is automatically determined.

    Parameters
    ----------
    filename : str
            path of the file

    client: dask.distributed.Client
            client object in distributed computation

    worker: dask.distributed.Worker
            worker object when running inside of a dask cluster

    **kwargs:
            additional keyword arguments for the underlying read functions

    Returns
    -------
    xarray.Dataset
    """

    # get the type of the file
    file_type = get_file_type(filename)
    if file_type is None:
        raise ValueError("unable to guess the type of the input file '%s'" % filename)

    # read the data
    if file_type in ["NC", "HDF"]:
        # use h5netcdf to read HDF5-Files
        engine = None
        if file_type == "HDF":
            engine = "h5netcdf"
        if kwargs.get("in_memory", False) and client is None:
            result0 = xarray.open_dataset(filename, engine=engine, decode_times=decode_times)
            result = result0.compute().chunk()
            result0.close()
            result.close()
        else:
            result = xarray.open_dataset(filename, engine=engine, decode_times=decode_times, chunks={}, **kwargs)
            if client is not None:
                if worker is not None:
                    logging.debug("running on worker: %s" % worker.address)
                    distributed.secede()
                    result = result.persist()
                    distributed.rejoin()
                else:
                    result = result.persist()
    elif file_type == "GRIB":
        result = read_grib_file(filename,
                                debug=kwargs.get("debug", False),
                                in_memory=kwargs.get("in_memory", False),
                                leadtime_from_filename=kwargs.get("leadtime_from_filename", False),
                                client=client,
                                worker=worker)
    else:
        raise ValueError("unknown file type '%s' for file '%s'" % (file_type, filename))

    # check for additional coordinate variables like staggered lat/lon values
    for one_name, one_var in six.iteritems(result.data_vars):
        if is_additional_coordinate_variable(one_var):
            try:
                result.set_coords(one_name, True)
            except TypeError:
                result.set_coords(one_name)

    # drop unused coords
    if kwargs.get("drop_unused", False):
        result = drop_unused(result)

    # add an ensemble dimension if we want to create one by folder or filename
    if kwargs.get("create_ens_dim", False):
        add_ensemble_dim(result, -99)
    return result


def expand_file_pattern(pattern):
    """
    use glob to find all files matching the pattern

    Parameters
    ----------
    pattern : str
            unix bash shell like search pattern

    Returns
    -------
    list
            list of file names matching the pattern
    """
    assert pattern is not None
    if "*" in pattern or "?" in pattern or "[" in pattern or "{" in pattern:
        files = glob.glob(pattern)
        return files
    else:
        return [pattern]


def __assign_ensemble_member_number_to_folders(folders):
    """
    used internally to extract numbers from the folders and use them as ensemble member numbers. If not all folders
    contain numbers, or if they are not unique, then create consecutive numbers.

    Parameters
    ----------
    folders : list
            list of absolute paths

    Returns
    -------
    dict :
            dictionary where the key is the absolute path and the value if the ensemble member number
    """

    # extract all digits from the last parts of the folders
    digits = []
    for ifolder, folder in enumerate(folders):
        prefix, one_path = os.path.split(folder)
        digits_for_path = ""
        for letter in one_path:
            if letter.isdigit():
                digits_for_path += letter
        if digits_for_path == "":
            digits_for_path = ifolder + 1
        else:
            digits_for_path = int(digits_for_path)
        digits.append(digits_for_path)

    # are all numbers unique?
    digits = np.unique(digits)
    if len(digits) < len(folders):
        digits = np.linspace(1, len(folders), len(folders), dtype=np.int32)

    # create the result dictionary
    result = {}
    for ifolder in range(len(folders)):
        result[folders[ifolder]] = digits[ifolder]
    return result



"""
different methods to display ensemble data
"""
import numpy as np
import xarray
from enstools.plot.core import get_nice_levels, get_coordinates_from_xarray


def grid(plot_function, variable, lon=None, lat=None, figure=None, axes=None, shape=None, cmaps=None, **kwargs):
    """
    Create a multi-panel plot by mapping the first dimension of the input array to a plot function specified in the
    second argument.

    Parameters
    ----------
    plot_function : Callable
            the plot function to use. supported are all function in enstools.plot

    variable : xarray.DataArray or np.ndarray
            the data to plot. Should be an 3d-Array, the first dimension is used to create the grid.

    lon : xarray.DataArray or np.ndarray or str
            longitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.

    lat : xarray.DataArray or np.ndarray or str
            latitude coordinate or name of longitude coordinate. The name may only be used for xarray variables.

    figure : matplotlib.figure.Figure
            If provided, this figure instance will be used (and returned), otherwise a new
            figure will be created.

    axes : array of matplotlib.axes.Axes
            if provided, this has to be an array of matplotlib.axes.Axes instances with the desired shape of the plot.

    shape : tuple
            The shape of the multi-panel plot (nrows, ncols). Not providing shape or axes is an error!

    cmaps : list
            A list of cmap names. The number of entries has to be the number of panel plots to create.

    **kwargs:
            Keyword arguments, they are forwarded top the actual plt function that creates the individual panels

    Returns
    -------
    tuple:
            fig, ax, where ax is an array for Axes objects

    Examples
    --------

    >>> fig, ax = grid(contour, data["TOT_PREC"][27, :, :, :], data["rlon"], data["rlat"], shape=(4, 5), cmaps=cmaps, rotated_pole=data["rotated_pole"], colorbar=False)    # doctest: +SKIP

    .. figure:: images/example_cluster_cosmo_01.png

        Example of displaying the result of K-Mean clustering for the COSMO-DE Ensemble.
        Each cluster has a different colormap.

    """
    # check arguments
    if variable.ndim != 3:
        raise ValueError("the input variable has to have three dimension, given: %d" % variable.ndim)
    if shape is None and axes is not None:
        shape = axes.shape
    if shape is None:
        raise ValueError("either shape or axes has to be provided!")
    if len(shape) != 2:
        raise ValueError("shape has to be 2d!")

    # create an array for the axes objects
    if axes is None:
        axes = np.empty(shape, dtype=object)

    # calculate levels for all panels
    if kwargs.get("levels", None) is None:
        kwargs["levels"] = get_nice_levels(variable)

    # assign color maps
    cmap = kwargs.get("cmap", None)
    if cmaps is None:
        cmaps = [cmap] * variable.shape[0]
    if len(cmaps) != variable.shape[0]:
        raise ValueError("wrong number of colormaps given!")

    # loop over all panels
    for ipanel in range(variable.shape[0]):
        figure, axes[np.unravel_index(ipanel, shape)] = plot_function(variable[ipanel, ...],
                                                                      lon,
                                                                      lat,
                                                                      figure=figure,
                                                                      subplot_args=(shape[0], shape[1], ipanel+1),
                                                                      cmap=cmaps[ipanel],
                                                                      **kwargs)
    return figure, axes



"""
internally used classes und functions that describe a batch job. This is meant for batchjobs inside of which this python
script is executed
"""
import os
import sys
from abc import ABCMeta, abstractmethod
import six
import traceback
import atexit
import socket
import distributed
import logging
from time import sleep
import shutil
import multiprocessing
import threading
import psutil
from .os_support import get_first_free_port, which, getenv, ProcessObserver, get_ip_address


@six.add_metaclass(ABCMeta)
class BatchJob():
    """
    abstract base class for a batch job
    """

    # scheduler and work processes started inside of this job.
    def __init__(self, local_dir=None):
        """
        read the environment in initialize some variables.
        """
        self.lock = threading.Lock()
        self.child_processes = []
        self.client = None
        self.cluster = None
        self.local_dir = local_dir
        self.ip_address = get_ip_address()
        atexit.register(self.cleanup)

    def start(self):
        """
        start the scheduler and the worker processes
        """
        # 1st step: start processes
        self.start_dask_scheduler()
        self.start_dask_worker()
        self.client = distributed.Client("tcp://%s:%d" % (self.ip_address, self.scheduler_port))

        # 2nd step: wait until all workers are started and connected to the scheduler
        while len(self.client.scheduler_info()["workers"]) < self.ntasks:
            logging.debug("waiting for workers... (%d/%d)" % (len(self.client.scheduler_info()["workers"]), self.ntasks))
            sleep(1)

        # 3rd step: make sure that all workers have the same PYTHONPATH
        def set_python_path(first_element):
            sys.path.insert(0, first_element)
        self.client.run(set_python_path, sys.path[0])
        self.client.run_on_scheduler(set_python_path, sys.path[0])

    def start_local(self):
        """
        start a local cluster. This can be used by different implementation as fallback
        """
        # TODO: the start procedure is not 100% reliable, change that!
        #       Possibly related: https://github.com/dask/distributed/issues/1321

        # set memory limit to 90% of the total system memory
        mem_per_worker = int(self.memory_per_node * 0.9) / self.ntasks
        logging.debug("memory limit per worker: %db" % mem_per_worker)

        # create the cluster by starting the client
        self.scheduler_port = get_first_free_port(self.ip_address, 8786)
        # When using a single node, the port and ip options won't be specified.
        """
        self.cluster = distributed.LocalCluster(n_workers=self.ntasks,
                                         local_dir=self.local_dir,
                                         scheduler_port=self.scheduler_port,
                                         ip=self.ip_address,
                                         silence_logs=logging.WARN,
                                         threads_per_worker=1,
                                         memory_limit=mem_per_worker)
        """
        self.cluster = distributed.LocalCluster(n_workers=self.ntasks,
                                         local_directory=self.local_dir,
                                         silence_logs=logging.WARN,
                                         threads_per_worker=1,
                                         memory_limit=mem_per_worker)
        self.client = distributed.Client(self.cluster)
        logging.debug("client and cluster started: %s" % str(self.client))

    def start_dask_scheduler(self):
        """
        the scheduler is started as child of this process
        """
        # get a free port for the new scheduler
        self.scheduler_port = get_first_free_port(self.ip_address, 8786)
        # TODO: check if successful!
        args = [which(sys.executable), which("dask-scheduler"), "--port", "%d" % self.scheduler_port, "--host", self.ip_address]
        logging.debug("Scheduler has been started at port %d and host %s" % (self.scheduler_port, self.ip_address))
        #if self.local_dir is not None:
        #    args.insert(2, "--local-directory")
        #    args.insert(3, self.local_dir)
        p = ProcessObserver(args)
        self.child_processes.append(p)

    @abstractmethod
    def start_dask_worker(self):
        """
        Start a worker for every SLURM-Task
        """
        pass

    def get_client(self):
        """
        if not already running, a new client is started and returned
        """
        return self.client

    def cleanup(self):
        """
        terminate all children
        """
        # cleanup is called more then once. The reason is, that it is called by __del__ in case of a normal exit and
        # also by atexit to ensure cleanup in case of a exit by exception.
        self.lock.acquire()
        try:
            # shutdown all workers
            if self.client is not None:
                # get a list of all workers
                workers = list(self.client.scheduler_info()['workers'])

                # if there is still something running on the cluster, cancel it
                data_refs = self.client.has_what()
                for one_worker, keys in six.iteritems(data_refs):
                    for key in keys:
                        f = distributed.Future(key, client=self.client)
                        self.client.cancel(f)

                # stop the workers
                self.client.sync(self.client.scheduler.retire_workers, workers=workers, close_workers=True)

                # wait for the workers to exit
                for one_child in self.child_processes[1:]:
                    for sec in range(15):
                        one_child.poll()
                        if one_child.returncode is not None:
                            break
                        else:
                            sleep(1)
                    if one_child.returncode is None:
                        logging.warning("At least one dask worker didn't quit voluntarily!")

                # close the connection to the scheduler
                self.client.close()
                self.client = None
        except:
            pass
        finally:
            try:
                # terminate remaining processes (should only be the scheduler)
                for one_child in self.child_processes:
                    if one_child.returncode is None:
                        one_child.terminate()

                # remove the temporal folder on the local host
                if self.local_dir is not None and os.path.exists(self.local_dir):
                    shutil.rmtree(self.local_dir)
                    self.local_dir = None
            except:
                pass
            finally:
                self.lock.release()

    def __del__(self):
        """
        remove children when this object is removed from memory.
        """
        self.cleanup()


class SlurmJob(BatchJob):
    """
    An interface to the SLURM-job inside of which we are running
    """
    def __init__(self, local_dir=None, ntasks=None):
        """
        read environment
        """
        super(SlurmJob, self).__init__(local_dir)
        self.job_id = getenv("SLURM_JOBID")
        self.step_id = os.getenv("SLURM_STEP_ID")
        self.ntasks = int(getenv("SLURM_NTASKS"))
        if ntasks is not None:
            if ntasks > self.ntasks:
                logging.warning("more worker requested than the number of tasks within this SLURM job.")
                logging.warning("number of worker automatically reduced to %d" % self.ntasks)
            else:
                self.ntasks = ntasks
        # if we are already inside of a slurm step, we can not start more steps. In this case,
        # do not use srun, but start a local cluster
        if self.step_id is not None:
            self.ip_address = "127.0.0.1"
            self.nnodes = 1
            self.nodelist = socket.gethostname()
        else:
            self.nnodes = int(getenv("SLURM_NNODES"))
            self.nodelist = getenv("SLURM_JOB_NODELIST")
        self.memory_per_node = int(getenv("SLURM_MEM_PER_NODE")) * 1048576    # unit: Byte

    def start(self):
        """
        Start a cluster using srun, but only if we are not inside of a slurm step
        """
        if self.step_id is not None:
            self.start_local()
        else:
            super(SlurmJob, self).start()

    def start_dask_worker(self):
        """
        use srun to start a worker on every allocated cpu.
        """
        # calculate the available memory per worker
        mem_per_worker = (self.memory_per_node * self.nnodes) // self.ntasks
        args = [which("srun"),
                "--ntasks=%d" % self.ntasks,
                sys.executable, which("dask-worker"),
                "--nthreads", "1",
                "--memory-limit", "%d" % mem_per_worker,
                "tcp://%s:%d" % (self.ip_address, self.scheduler_port)
                ]
        logging.debug(" ".join(args))
        #if self.local_dir is not None:
        #    args.insert(4, "--local-directory")
        #    args.insert(5, self.local_dir)
        p = ProcessObserver(args)  # subprocess.Popen(args)
        self.child_processes.append(p)

        # specify a cleanup command to execute after the workers finished
        if self.local_dir is not None:
            p.run_on_exit([which("srun"), "--ntasks=%d" % self.nnodes, "--ntasks-per-node=1", which("rm"), "-rf", self.local_dir])


class LocalJob(BatchJob):
    """
    start a LocalCluster instead of a batch job cluster
    """
    def __init__(self, local_dir=None, ntasks=None):
        super(LocalJob, self).__init__(local_dir)
        self.ntasks = _get_num_available_procs()
        if ntasks is not None:
            if ntasks > self.ntasks:
                logging.warning("More worker requested then CPUs available.")
                logging.warning("Since we are running on a local computer without job scheduler, they are started anyway.")
            self.ntasks = ntasks
        self.ip_address = "127.0.0.1"
        self.nnodes = 1
        self.nodelist = socket.gethostname()
        self.memory_per_node = psutil.virtual_memory().total

    def start_dask_worker(self):
        pass

    def start(self):
        """
        Start a local cluster
        """
        self.start_local()

    def cleanup(self):
        self.lock.acquire()
        try:
            # close the cluster and the connection to the cluster
            if self.client is not None:
                # if there is still something running on the cluster, cancel it
                data_refs = self.client.has_what()
                for one_worker, keys in six.iteritems(data_refs):
                    for key in keys:
                        f = distributed.Future(key, client=self.client)
                        self.client.cancel(f)

                # close the client and the cluster itself
                self.client.close()
                self.client = None
                workers = list(self.cluster.workers)
                try:
                    for one_worker in workers:
                        self.cluster.stop_worker(one_worker)
                except AttributeError:
                    pass
                self.cluster.close()

            # remove the temporal folder on the local host
            if self.local_dir is not None and os.path.exists(self.local_dir):
                shutil.rmtree(self.local_dir)
                self.local_dir = None
        except:
            raise
        finally:
            self.lock.release()


def get_batch_job(local_dir=None, **kwargs):
    """
    read several environment variables to figure out whether or not we are inside of a Job script or not.

    Parameters
    ----------
    local_dir : str
            absolute path of a folder used as temporal directory

    **kwargs:
            all arguments are passed on the the init function of the different job implementations

    Returns
    -------
    BatchJob :
            A batch job object for the detected Job type
    """
    # are we inside of SLURM?
    job_id = os.getenv("SLURM_JOB_ID")
    if job_id is not None:
        # Check if we are in more than one node:
        node_list = os.getenv("SLURM_NODELIST")
        logging.debug("Node list: %s" % node_list)
        number_of_nodes = int(os.getenv("SLURM_NNODES"))
        logging.debug("Number of nodes in the current slurm job: %d" % number_of_nodes)
        if number_of_nodes > 1:
            
            return SlurmJob(local_dir, **kwargs)

    # no job environment detected? create a local cluster!
    return LocalJob(local_dir, **kwargs)

    # no scheduler found?
    raise ValueError("unable to create a new dask cluster!")


def _get_num_available_procs():
    """
    Find the number of processors available for computations. The function will look for environment variables from
    SLURM and openMP (in that order). If no environment variables are defined, the number of cpus will be returned.

    Returns
    -------
    int
            Number of processes available for parallel computations.
    """
    # running inside a slurm job?
    SLURM_NTASKS = os.getenv("SLURM_NTASKS")
    if SLURM_NTASKS is not None:
        return int(SLURM_NTASKS)

    # openmp number of threads?
    OMP_NUM_THREADS = os.getenv("OMP_NUM_THREADS")
    if OMP_NUM_THREADS is not None:
        return int(OMP_NUM_THREADS)

    # no hints from environment variables, the number of hardware cpus is returned
    return multiprocessing.cpu_count()



import numpy as np
import xarray
from .fix_attributes import fix_attributes
from enstools.core.errors import EnstoolsError


def continuous_ranked_probability_score(reference: xarray.DataArray, target: xarray.DataArray,
                                        ensemble_dimension: str = "ens") -> xarray.DataArray:
    """
    Description: My implementation of the Continuous Ranked Probability Score.

    Functions
    :param reference:
    :param target:
    :param ensemble_dimension: name of the ensemble dimension, usually is 'ens' or 'member'
    :return:
    """
    if ensemble_dimension not in reference.dims:
        raise EnstoolsError(f"Trying to compute CRPS on a dataset that doesn't contain ensemble_dimension:{ensemble_dimension}")
    # Make a copy of the reference data array to store the results in it.
    result = reference.copy(deep=True)
    result = result.sel(ens=1)

    # Reorder coordinates to put the ensemble dimension in first place
    reference = reference.transpose(ensemble_dimension, ...)
    target = target.transpose(ensemble_dimension, ...)

    # Get ensemble sizes
    reference_ensemble_size = reference[ensemble_dimension].size
    target_ensemble_size = target[ensemble_dimension].size

    ensemble_index = reference.dims.index(ensemble_dimension)

    # We need to find the minimum and maximum values for each grid point for both datasets
    ref_min, ref_max = reference.min(dim=ensemble_dimension), reference.max(dim=ensemble_dimension)
    trg_min, trg_max = target.min(dim=ensemble_dimension), target.max(dim=ensemble_dimension)

    # Get minimums and maximums for each grid point.
    _min = np.stack((ref_min, trg_min)).min(axis=0)
    _max = np.stack((ref_max, trg_max)).max(axis=0)

    # For each grid point, create a linear space that will cover the full range of values of the two arrays
    x = np.linspace(_min, _max, reference_ensemble_size)

    # Two empty arrays are created, will be used to count the number of data values that are below certain threshold
    # to estimate the ECDF in both arrays using the same reference
    value_occurrence_ref = np.zeros(reference.shape)
    value_occurrence_trg = np.zeros(reference.shape)

    # Load the data to avoid problems with dask
    reference.load()
    target.load()


    for member in range(reference["ens"].size):
        selection = tuple([member if dim == ensemble_dimension else slice(None) for dim in reference.dims])
        value_occurrence_ref[selection] = (reference < x[member]).sum(dim=ensemble_dimension)
        value_occurrence_trg[selection] = (target < x[member]).sum(dim=ensemble_dimension)

    # We normalize by the size of each ensemble
    value_occurrence_ref = value_occurrence_ref / reference_ensemble_size
    value_occurrence_trg = value_occurrence_trg / target_ensemble_size

    # Compute error as the area between the ECDF
    area = np.trapz(np.abs(value_occurrence_ref - value_occurrence_trg), x, axis=ensemble_index)

    result.values = area

    # Add crps to attributes and delete obsolete ones:
    result = fix_attributes(result, suffix="crps")

    # Return error average
    return result



import numpy as np
import xarray
from enstools.core import check_arguments
from enstools.misc import distance
from numba import njit


@check_arguments(dims={'u': ('lat', 'lon'), 'v': ('lat', 'lon')}, shape={'lon': (0,), 'lat': (0,)})
def vorticity(u, v, lon, lat, fill_value=np.NaN):
    """
    Calculate relative vorticity and its components shear and curvature on a regular lat-lon-grid.

    Parameters
    ----------
    u: xarray.DataArray
        u-component of the wind field

    v: xarray.DataArray
        v-component of the wind field

    lon: xarray.DataArray or np.ndarray
        longitude of the input data in degrees. 1D-Array

    lat: xarray.DataArray or np.ndarray
        latitude of the input data in degrees. 1D-Array

    fill_value: float
        fill value for points at the border. Default: NaN.

    Returns
    -------
    vorticity, shear_vorticity, curve_vorticity: xarray.DataArray
    """
    # perform the actual calculation with numba
    vor, shear, curve = __vorticity(np.asarray(u), np.asarray(v), np.asarray(lon), np.asarray(lat), fill_value)

    # convert result in xarray.DataArrays
    vor = xarray.DataArray(
        vor, coords=[lat, lon], dims=('lat', 'lon'),
        name="relative_vorticity",
        attrs={
            "standard_name": "atmosphere_upward_relative_vorticity",
            "units": "s-1"
        }
    )
    shear = xarray.DataArray(
        shear, coords=[lat, lon], dims=('lat', 'lon'),
        name="shear_vorticity",
        attrs={
            "standard_name": "shear_component_of_atmosphere_upward_relative_vorticity",
            "units": "s-1"
        }
    )
    curve = xarray.DataArray(
        curve, coords=[lat, lon], dims=('lat', 'lon'),
        name="curve_vorticity",
        attrs={
            "standard_name": "curvature_component_atmosphere_upward_relative_vorticity",
            "units": "s-1"
        }
    )
    return vor, shear, curve


@njit()
def __vorticity(u, v, lon, lat, fill_value):
    vor = np.zeros(u.shape, dtype="float")
    shear = np.zeros(u.shape, dtype="float")
    curve = np.zeros(u.shape, dtype="float")
    v1 = np.zeros(2)
    v2 = np.zeros(2)

    # loop over all grid points
    for y in range(1, u.shape[0]-1):
        for x in range(1, u.shape[1]-1):
            # any missing values?
            if np.isnan(u[y,x]) or np.isnan(u[y,x+1]) or np.isnan(u[y,x-1]) \
                or np.isnan(u[y+1,x]) or np.isnan(u[y-1,x]) \
                or np.isnan(v[y,x]) or np.isnan(v[y,x+1]) or np.isnan(v[y,x-1]) \
                or np.isnan(v[y+1,x]) or np.isnan(v[y-1,x]):
                vor[y,x] = np.NaN
                shear[y,x] = np.NaN
                curve[y,x] = np.NaN
            
            # distance on globe
            dx = distance(lat[y], lat[y], lon[x+1], lon[x-1], input_in_radian=False)
            dy = distance(lat[y-1], lat[y+1], lon[x], lon[x], input_in_radian=False)
            vor[y,x] = (v[y,x+1]-v[y,x-1]) / dx - (u[y+1,x]-u[y-1,x]) / dy

            # calculate shear-vorticity
            # calculate the wind direction
            wdir = np.arctan2(u[y,x], v[y,x])
            wspd = np.sqrt(u[y,x]**2 + v[y,x]**2)
            sin_wdir = np.sin(wdir)
            cos_wdir = np.cos(wdir)

            # calculate dot-product for four points around the reference point to the reference vector.
            # Use the wind component parallel to the reference vector from all four points and calculate
            # the vorticity based on this component, which results in the shear vorticity.
            #
            # This approach follows Berry et al. 2006.
            # Points:
            #    c
            #  b r a
            #    d

            # point a
            v1[0] = u[y,x]
            v1[1] = v[y,x]
            v2[0] = u[y,x+1]
            v2[1] = v[y,x+1]
            dp = np.dot(v1, v2) / wspd
            v_a = dp * cos_wdir

            # point b
            v2[0] = u[y,x-1]
            v2[1] = v[y,x-1]
            dp = np.dot(v1, v2) / wspd
            v_b = dp * cos_wdir

            # point c
            v2[0] = u[y+1,x]
            v2[1] = v[y+1,x]
            dp = np.dot(v1, v2) / wspd
            u_c = dp * sin_wdir

            # point d 
            v2[0] = u[y-1,x]
            v2[1] = v[y-1,x]
            dp = np.dot(v1, v2) / wspd
            u_d = dp * sin_wdir

            # calculate the shear vorticity
            shear[y,x] = (v_a-v_b) / dx - (u_c-u_d) / dy
    
    # fill values at the borders
    vor[0,:] = np.NaN
    vor[-1,:] = np.NaN
    vor[:,0] = np.NaN
    vor[:,-1] = np.NaN
    
    shear[0,:] = np.NaN
    shear[-1,:] = np.NaN
    shear[:,0] = np.NaN
    shear[:,-1] = np.NaN
    
    # calculate curvature-vorticity 
    curve = np.where(np.logical_or(np.isnan(vor), np.isnan(shear)), np.NaN, vor-shear)

    # replace fill values if something different from NaN is used
    if not np.isnan(fill_value):
        vor = np.where(np.isnan(vor), fill_value, vor)
        shear = np.where(np.isnan(shear), fill_value, shear)
        curve = np.where(np.isnan(curve), fill_value, curve)
    return vor, shear, curve



"""
Command line interface for enstools.opendata. Currently only data from DWD is available.
"""
import os
import argparse
import logging
from datetime import datetime
from enstools.opendata import retrieve_nwp, getDWDContent

# default format used on the command line.
__DATE_FORMAT="%Y-%m-%dT%H:%M"


def error_exit(text):
    """
    quit with error code 1

    Parameters
    ----------
    text: str
            error message to show before quiting.
    """
    logging.error(text)
    exit(1)


def retrieve_data(args):
    """
    use enstools.opendata.retrieve_nwp to download files

    Parameters
    ----------
    args:
            command-line arguments
    """
    if args.model is None:
        error_exit("--model is required for retrieve!")
    if args.grid_type is None:
        error_exit("--grid_type is required for retrieve!")
    if args.init_time is None:
        error_exit("--init-time is required for retrieve!")
    if args.variable is None:
        error_exit("--variable is required for retrieve!")
    if args.level_type is None:
        error_exit("--level-type is required for retrieve!")
    if args.levels is None:
        error_exit("--levels is required for retrieve!")
    if args.lead_time is None:
        error_exit("--lead-time is required for retrieve!")
    if args.dest is None:
        error_exit("--dest is required for retrieve!")
    if not os.path.exists(args.dest):
        error_exit(f"path not found: {args.dest}")
    if not os.path.isdir(args.dest):
        error_exit(f"path is not a directory: {args.dest}")
    try:
        files = retrieve_nwp(model=args.model,
                             grid_type=args.grid_type,
                             variable=args.variable,
                             level_type=args.level_type,
                             levels=args.levels,
                             init_time=args.init_time,
                             forecast_hour=args.lead_time,
                             dest=args.dest,
                             merge_files=args.merge)
        for one_file in files:
            logging.info(f"downloaded: {one_file}")
    except ValueError as ve:
        error_exit(str(ve))

def query(args):
    """
    find files on opendata platforms

    Parameters
    ----------
    args:
            command-line arguments
    """
    content = getDWDContent()

    # list all available models
    if args.get_models:
        logging.info(f"Available models: {', '.join(content.get_models())}")

    # the available init times for a specific model/grid
    if args.get_init_times:
        if args.model is None:
            error_exit("--model is required for this query!")
        try:
            init_times = content.get_avail_init_times(args.model, args.grid_type)
            init_times_str = []
            for one_time in init_times:
                if type(one_time) == int:
                    init_times_str.append(str(one_time))
                elif isinstance(one_time, datetime):
                    init_times_str.append(one_time.strftime(__DATE_FORMAT))
            logging.info(f"Available init times: {', '.join(init_times_str)}")
        except ValueError as ve:
            error_exit(str(ve))

    # the available variables for a model, grid and init time
    if args.get_vars:
        if args.model is None:
            error_exit("--model is required for this query!")
        if args.init_time is None:
            error_exit("--init-time is required for this query!")
        try:
            logging.info(f"Available variables: {', '.join(content.get_avail_vars(model=args.model, grid_type=args.grid_type, init_time=args.init_time))}")
        except ValueError as ve:
            error_exit(str(ve))

    if args.get_level_types:
        if args.model is None:
            error_exit("--model is required for this query!")
        if args.init_time is None:
            error_exit("--init-time is required for this query!")
        if args.variable is None:
            error_exit("--variable is required for this query!")
        try:
            logging.info(f"Available level-types: {', '.join(content.get_avail_level_types(model=args.model, grid_type=args.grid_type, init_time=args.init_time, variable=args.variable[0]))}")
        except ValueError as ve:
            error_exit(str(ve))

    if args.get_levels:
        if args.model is None:
            error_exit("--model is required for this query!")
        if args.init_time is None:
            error_exit("--init-time is required for this query!")
        if args.variable is None:
            error_exit("--variable is required for this query!")
        if args.level_type is None:
            error_exit("--level-type is required for this query!")
        try:
            levels = content.get_avail_levels(model=args.model, grid_type=args.grid_type, init_time=args.init_time, variable=args.variable[0], level_type=args.level_type)
            levels = list(map(lambda x: str(x), levels))
            logging.info(f"Available levels: {', '.join(levels)}")
        except ValueError as ve:
            error_exit(str(ve))

    if args.get_lead_times:
        if args.model is None:
            error_exit("--model is required for this query!")
        if args.init_time is None:
            error_exit("--init-time is required for this query!")
        if args.variable is None:
            error_exit("--variable is required for this query!")
        if args.level_type is None:
            error_exit("--level-type is required for this query!")
        try:
            hours = content.get_avail_forecast_hours(model=args.model, grid_type=args.grid_type, init_time=args.init_time, variable=args.variable[0], level_type=args.level_type)
            hours = list(map(lambda x: f"{x}h", hours))
            logging.info(f"Available forecast lead-times: {', '.join(hours)}")
        except ValueError as ve:
            error_exit(str(ve))


def __add_common_args(parser: argparse.ArgumentParser):
    """
    add common arguments to sub-parsers
    """
    parser.add_argument("--model", help="name of the model to use. Use query --get-models to get a list of valid names.")
    parser.add_argument("--grid-type", help="type of the grid to use.")
    parser.add_argument("--level-type", help="type of the vertical level to use.")
    parser.add_argument("--init-time", help=f"initialization time to use. "
                                            "Integers are interpreted as hours since model start, dates formatted as "
                                            f"{__DATE_FORMAT.replace('%Y', 'YYYY').replace('%m', 'MM').replace('%d', 'DD').replace('%H', 'HH').replace('%M', 'MM')} are interpreted as absolute start dates.")
    parser.add_argument("--variable", nargs="+", help="name of the variable to use. Use query --get-vars to get a list of valid names.")
    parser.add_argument("--levels", nargs="+", type=int, help="levels to use.")
    parser.add_argument("--lead-time", nargs="+", type=int, help="lead times to use in hours.")


def __parse_init_time(args):
    """
    is the command line arguments contain an init time, then it is here converted to datetime
    """
    if args.init_time is None:
        return
    try:
        if args.init_time.isdigit():
            args.init_time=int(args.init_time)
        else:
            args.init_time=datetime.strptime(args.init_time, __DATE_FORMAT)
    except Exception as ex:
        error_exit(str(ex))


def main():
    # parse command line arguments
    parser = argparse.ArgumentParser(description=__doc__)
    subparsers = parser.add_subparsers(help="Searching and downloading data is available as sub-commands.")

    # sub-command for downloading data
    parser_retrieve = subparsers.add_parser("retrieve", help="download files from an opendata service provider")
    parser_retrieve.add_argument("--dest", help="destination folder for downloaded data.")
    parser_retrieve.add_argument("--merge", action='store_true', help="merge multiple downloaded files.")
    __add_common_args(parser_retrieve)
    parser_retrieve.set_defaults(func=retrieve_data)

    # sub-command for finding data
    parser_query = subparsers.add_parser("query", help="search for data on opendata service providers")
    parser_query.add_argument("--get-models", action='store_true', help="list available models.")
    parser_query.add_argument("--get-init-times", action='store_true', help="list available init times. Requires argument --model and accepts --grid-type.")
    parser_query.add_argument("--get-vars", action='store_true', help="list of available variables for --model, --grid-type, and --init-time.")
    parser_query.add_argument("--get-level-types", action='store_true', help="get the available types of vertical levels for --model, --grid-type, --init-time, and --variable.")
    parser_query.add_argument("--get-levels", action='store_true', help="get the available vertical levels for --model, --grid-type, --init-time, --variable, and --level-type.")
    parser_query.add_argument("--get-lead-times", action='store_true', help="get the available times since model start for --model, --grid-type, --init-time, and --level-type.")
    __add_common_args(parser_query)
    parser_query.set_defaults(func=query)

    args = parser.parse_args()
    __parse_init_time(args)
    args.func(args)



"""
Wrapper around sklearn cluster methods for more convenience.
"""
import sklearn.cluster
from sklearn.metrics import silhouette_score
import numpy as np
import xarray
import dask.delayed
import dask.array

methods = {"kmeans": sklearn.cluster.KMeans,
           "aprop": sklearn.cluster.AffinityPropagation,
           "mshift": sklearn.cluster.MeanShift,
           "spectral": sklearn.cluster.SpectralClustering,
           "agglo": sklearn.cluster.AgglomerativeClustering,
           "dbscan": sklearn.cluster.DBSCAN,
           "birch": sklearn.cluster.Birch}

methods_with_ncluster = {
           "kmeans": True,
           "aprop": False,
           "mshift": False,
           "spectral": True,
           "agglo": True,
           "dbscan": False,
           "birch": True}


def cluster(algorithm, data, n_clusters=None, n_clusters_max=None, sort=True, **kwargs):
    """

    Parameters
    ----------
    algorithm : str
            clustering method to use. all sklearn methods are supported. $methods

    data : xarray.DataArray or np.ndarray
            the input data for the clustering. This is the output of the :func:`prepare` function.

    n_clusters : int or None
            number of clusters to create. An Integer will create the specified number of
            clusters. None will try to estimate the number of clusters using the silhouette score if the algorithm
            supports the prescription of the number of clusters.

    n_clusters_max : int
            maximal number of clusters to create. The default is the number of ensemble members divided by four.

    sort : bool
            If True, the clusters are sorted in order to create better reproducible results.

    **kwargs
            all

    Returns
    -------
    np.ndarray
            1d array with cluster labels.
    """

    # is a valid algorithm selected?
    if algorithm not in methods:
        raise ValueError("unsupported algorithm selected: %s supported are only: %s" % (algorithm, ", ".join(methods.keys())))

    # is it possible to prescribe the number of clusters?
    has_n_clusters = methods_with_ncluster[algorithm]

    # estimate the number of clusters?
    scores = None
    if has_n_clusters and n_clusters is None:
        # calculate clustering up the 1/3 times the ensemble members
        n_clusters_min = 2
        if n_clusters_max is None:
            n_clusters_max = data.shape[0] // 4

        labels = []
        scores = []
        for nc in range(n_clusters_min, n_clusters_max + 1):
            labels.append(dask.delayed(cluster)(algorithm, data, n_clusters=nc, sort=False, **kwargs))
        labels = dask.compute(*labels)
        for nc in range(n_clusters_min, n_clusters_max + 1):
            scores.append(dask.delayed(silhouette_score)(data, np.asarray(labels[nc - n_clusters_min])))

        # the the number of clusters with the largest score
        scores = np.array(dask.compute(*scores))
        n_clusters = scores.argmax() + n_clusters_min
        result = labels[scores.argmax()]

    # number of clusters is given as argument
    elif has_n_clusters:
        n_clusters_min = n_clusters
        n_clusters_max = n_clusters
        model = methods[algorithm](n_clusters=n_clusters, **kwargs).fit(data)
        result = model.labels_

    # sort the result?
    if sort:
        # calculate the variance for each cluster
        cluster_in_res = np.unique(result)
        cluster_var = list(map(lambda x: data[np.asarray(result) == x, ...].var(axis=0).mean(), cluster_in_res))

        # sort the clusters by their variance between the members
        sorted_clusters = sorted(zip(cluster_in_res, cluster_var), key=lambda x: x[1])

        # relabel the result
        for i_cluster, one_cluster in enumerate(sorted_clusters):
            result = np.where(result == one_cluster[0], i_cluster + len(cluster_in_res), result)
        result -= len(cluster_in_res)

    # convert to xarray and add some meta data
    result = xarray.DataArray(result, dims=("ens",))
    result.attrs["n_clusters"] = int(result.max() + 1)
    if has_n_clusters:
        result.attrs["n_clusters_min"] = n_clusters_min
        result.attrs["n_clusters_max"] = n_clusters_max
    result.attrs["algorithm"] = algorithm
    if scores is not None:
        result.attrs["scores"] = scores

    return result


# add a list of methods to the doc-string
__methods_str = "\n\n"
for __method_name in sorted(methods.keys()):
    __methods_str += "            *%s*: :class:`sklearn.cluster.%s`\n\n" % (__method_name, methods[__method_name].__name__)
cluster.__doc__ = cluster.__doc__.replace("$methods", __methods_str)



from .DWDContent import getDWDContent


def retrieve_nwp(service="DWD", model="ICON", eps=None, grid_type=None, variable=None, level_type=None,
                 levels=0, init_time=None, forecast_hour=None, merge_files=False, dest=None):
    """
    Downloads numerical weather prediction (NWP) datasets from opendata server.
    Parameters
    ----------
    service : str
            name of weather service. Default="DWD".
    model : str
            name of the model. Default="ICON".

    eps : bool
            if True, download ensemble forecast, otherwise download deterministic forecast.

    grid_type: str
        The type of the geo grid.

    variable : list or str
            list of variables to download. Multiple values are allowed.

    level_type : str
            one of "model", "pressure", or "single"

    levels : list or int
            levels to download. Unit depends on `level_type`.

    init_time : int or str

    forecast_hour : list or int
            hours since the initialization of forecast. Multiple values are allowed.

    merge_files : bool
            if true, GRIB files are concatenated to create one file.

    dest : str
            Destination folder for downloaded data. If the files are already available,
            they are not downloaded again.

    Returns
    -------
    list :
            names of downloaded files.
    """
    content = getDWDContent()
    download_files = content.retrieve(service=service, model=model, eps=eps, grid_type=grid_type,
                                               variable=variable, level_type=level_type, levels=levels,
                                               init_time=init_time, forecast_hour=forecast_hour,
                                               merge_files=merge_files, dest=dest)
    return download_files



from .backend import InteractiveBackend
from .backend import Function, Stream

from enstools.misc import generate_coordinates
from enstools.interpolation import nearest_neighbour

from numba import jit
import numpy as np
from math import sqrt, floor
from collections import OrderedDict
from scipy.interpolate import interp2d

import cartopy
from matplotlib import cm
import matplotlib.pyplot as plt
from matplotlib.colors import Normalize
import plotly.graph_objs as go



class InteractiveBackend3D(InteractiveBackend):
    """
    Abstract backend class that will be implemented for different plotting backends. Functions that are used for all
    backends can be implemented directly here.
    """
    
    @classmethod
    def spherical_to_cartesian(cls, lons, lats, radius = 6371229.0):
        """
        Transforms spherical coordinates into 3D cartesian coordinates.
    
        Parameters
        ----------
        lons, lats : Numpy array
            Numpy arrays contain longitudes or latitudes in radians.
                 
        radius : float, optional, default=6371229.0
            Radius of a sphere for a transformation. For default an argument equals to Earth radius.
                
        Returns
        -------
        Numpy array
            3-dimensional Numpy array contains cartesian coordinates.
        """
            
        cartesian = np.empty((lons.shape[0], 3))
        lon = lons + np.pi
        lat = lats - np.pi / 2.0
        cartesian[:, 0] = radius * np.sin(lat) * np.cos(lon)
        cartesian[:, 1] = radius * np.sin(lat) * np.sin(lon)
        cartesian[:, 2] = radius * np.cos(lat)

        return cartesian
    
    @classmethod
    def separate_cartesian(cls, cartesian):  
        """
        Separates cartesian coordinates into X, Y and Z components.

        Parameters
        ----------
        cartesian: array
            3D array contains cartesian coordinates.
                      
        Returns
        -------
        xs, ys, zs : list of lists
            Lists of lists contain X-, Y- or Z- component of cartesian coordinates in radians.    
        """
        
        # separate X-, Y- and Z-coordinates
        xs = []
        ys = []
        zs = []
        for i in range(len(cartesian)):
            x_i = []
            y_i = []
            z_i = []
            for j in range(len(cartesian[i])):
                x_i.append(cartesian[i][j][0])
                y_i.append(cartesian[i][j][1])
                z_i.append(cartesian[i][j][2])
            xs.append(x_i)
            ys.append(y_i)
            zs.append(z_i)   
                                
        return xs, ys, zs
       
    @classmethod
    def paths_cartesian(cls, xs, ys, radius = 6381229.0):
        """
        Transforms paths of lines in degrees into paths in Cartesian coordinates leaving initial paths structure.

        Parameters
        ----------
        xs, ys : list of lists
            Lists of lists contain X- or Y-coordinates of paths.
            
        radius : float, optional, default=6371229.0
            Radius of a sphere for a transformation. For default an argument equals to Earth radius.          
                      
        Returns
        -------
        xs, ys, zs : list of lists
            Lists of lists contain X-, Y- or Z- component of cartesian coordinates of paths.                
        """
               
        # transform to cartesian coordinates
        xs_radians = np.array([np.radians(np.array(xi)) for xi in xs], dtype=object)
        ys_radians = np.array([np.radians(np.array(yi)) for yi in ys], dtype=object)
        
        cartesian = np.array([cls.spherical_to_cartesian(xs_radians[i], ys_radians[i], radius=radius) 
                              for i in np.arange(len(xs_radians))], dtype=object)     
        npaths = len(cartesian)
        
        # separate X-, Y- and Z-coordinates
        xs, ys, zs = cls.separate_cartesian(cartesian)  
                    
        return xs, ys, zs, npaths
    
    @staticmethod
    @jit(nopython=True)
    def point_in_region(lons, lats, x, y, cell_size):
        """
        Checks if point is within a region.
        
        Parameters
        ----------                           
        lons, lats : Numpy array
            Numpy arrays contain longitudes or latitudes of cell vertices in radians.
            
        x, y : float
            Values of longitude or latitude of the point in degrees. 
            
        cell_size : float
            Approximate size of a grid cell.

        Returns
        -------
        bool : 
            Equals to True if the point is within the region.
        """
        
        lons = np.degrees(lons)
        lats = np.degrees(lats)
        
        degrees_to_meters = 2 * np.pi * 6371229.0 / 360
        for i in range(len(lons)):
            in_region = False
            distance = (lons[i] - x)*(lons[i] - x) + (lats[i] - y)*(lats[i] - y)
            if degrees_to_meters*sqrt(distance) <= cell_size:
                in_region = True
                break
                
        return in_region

    @classmethod
    def make_path_map3D(cls, feature, cell_size, lons, lats, regional, lon0, lon1, lat0, lat1, map_resolution):
        """
        Defines paths of coastlines or country borders for a defined region. Includes selection of a defined region. 
        Adopted for a 3D plotting.

        Parameters
        ----------
        feature : {'coastlines', 'borders'}
            Argument 'coastlines' for paths of coastlines. Argument 'borders' for paths of country borders.
            
        cell_size : float
            Approximate size of a grid cell.
                
        lons, lats : Numpy array
            Numpy arrays contain longitudes or latitudes in radians.
            
        regional : bool
            Argument defines necessity to select part of paths for a defined region. Equals to True for a regional plot.
            
        lon0, lon1, lat0, lat1 : float
            Values of longitudes or latitudes of a defined region borders. 
            
        map_resolution : {'110m', '50m', '10m'}
            Definition of a map resolution.

        Returns
        -------
        xs_map, ys_map, zs_map : list of lists
            Lists of lists contain separated X-, Y- or Z-coordinates of paths in a defined region.
            
        npaths : int
            Number of paths
        """
        
        paths = []
        # create paths
        if feature == 'coastlines':
            for geom in cartopy.feature.COASTLINE.with_scale(map_resolution).geometries():
                if geom.geom_type == 'LineString':
                    paths.append(list(geom.coords))
                else:
                    for g in geom:
                        paths.append(list(g.coords))
                    
        elif feature == 'borders':
            for geom in cartopy.feature.BORDERS.with_scale(map_resolution).geometries():
                if geom.geom_type == 'LineString':
                    paths.append(list(geom.coords))
                else:
                    for g in geom:
                        paths.append(list(g.coords))
                    
        else:
            raise ValueError(f"unsupported feature: {feature}")
        
        # select paths which are in a defined region and separate X- and Y-coordinates of paths  
        # preselect paths with are in a square frame
        xs_frame = []
        ys_frame = []        
        for i in range(len(paths)):
            xs_i = []
            ys_i = []
            in_frame = True
            prev_step_in_frame = True
            for j in range(len(paths[i])):
                in_frame = paths[i][j][0]>=lon0 and paths[i][j][0]<=lon1 and paths[i][j][1]>=lat0 and paths[i][j][1]<=lat1
                if prev_step_in_frame:
                    if in_frame:
                        xs_i.append(paths[i][j][0])
                        ys_i.append(paths[i][j][1])
                        if j == len(paths[i])-1:
                            xs_frame.append(xs_i)
                            ys_frame.append(ys_i)
                    elif len(xs_i) != 0:
                        xs_frame.append(xs_i)
                        ys_frame.append(ys_i)
                elif in_frame:
                    xs_i = []
                    ys_i = []
                    xs_i.append(paths[i][j][0])
                    ys_i.append(paths[i][j][1])
                prev_step_in_frame = in_frame
        
        # final (more accurate) selection and transformation to degrees
        if regional:
            xs_degrees = []
            ys_degrees = []  
        
            for i in range(len(xs_frame)):
                xs_i = []
                ys_i = []
                for j in range(len(xs_frame[i])):
                    in_region = cls.point_in_region(lons, lats, xs_frame[i][j], ys_frame[i][j], cell_size)
                    if in_region:
                        xs_i.append(xs_frame[i][j])
                        ys_i.append(ys_frame[i][j])
                    elif len(xs_i) != 0:
                        xs_degrees.append(xs_i)
                        ys_degrees.append(ys_i)
                        xs_i = []
                        ys_i = [] 
                if in_region and len(xs_i) != 0:
                    xs_degrees.append(xs_i)
                    ys_degrees.append(ys_i)
        else:
            xs_degrees = xs_frame
            ys_degrees = ys_frame
        
        # transform to cartesian coordinates and separate coordinates
        xs_map, ys_map, zs_map, npaths = cls.paths_cartesian(xs_degrees, ys_degrees, radius=6381229.0)
        
        return xs_map, ys_map, zs_map, npaths
    
    @classmethod
    def set_aspectratio3D(cls, xs, ys, zs):
        """
        Defines an aspectratio of a figure based ob size of a region.

        Parameters
        ----------            
        xs, ys, zs : float
            Values of X-, Y- or Z-coordinates of a defined region. 

        Returns
        -------
        xratio, yratio, zratio
            Aspectratio for a 3D figure.
        """    
                  
        # get exices ranges          
        dx = abs(np.max(xs) - np.min(xs))
        dy = abs(np.max(ys) - np.min(ys))
        dz = abs(np.max(zs) - np.min(zs))
        norma = np.max(np.array([dx, dy, dz]))
        xratio = dx / norma
        yratio = dy / norma
        zratio = dz / norma
        
        return xratio, yratio, zratio  
    
    @classmethod    
    def set_colors(cls, data, colormap, function):
        """
        Defines colors of individual grid cells.
    
        Parameters
        ----------
        data : xarray
            Values of a variable and a range of other parameters describing a variable and a grid.
                 
        colormap : str
            Name of a colormap.
        
        function : Enum element
            Definition of a Plotly function used for a plotting. Element from Function Enum.

        Returns
        -------
        list
            List contains colors for every grid cell based on a defined colormap. 
        """
        
        # define colormap
        cmap = cm.get_cmap(colormap)
        # normalize values of a variable
        normalization = Normalize(vmin=np.min(data[~np.isnan(data)]), vmax=np.max(data[~np.isnan(data)]))
        normalized_data = normalization(data)
        # set colors 
        data_cmap = cmap(normalized_data)
        
        # create list of colors in a right format for plotting
        colors = []
        for i in range(len(data_cmap)):
            if function == Function.TRISURF:
                if data_cmap[i][0] == data_cmap[i][1] == data_cmap[i][2] == data_cmap[i][3] == 0:
                    colors.append(tuple((255,255,255,0)))
                else:
                    colors.append(tuple(data_cmap[i]))
            elif function == Function.MESH3D:
                colors.append(data_cmap[i])

        return colors
    
    @classmethod
    def make_colorbar(cls, colorset, valueset, tick_interval, variable_name, x_colorbar):
        """
        Creates a colorbar for Plotly.

        Parameters
        ----------
        colorset : list of tuples
            Set of ordered unique colors. List containing tuples mapping a normalized value to a HEX or RGB. 
                 
        valueset : Numpy array
            Set of ordered unique double values.
                 
        tick_interval : float
            Value of intervals between ticks in a colorbar.
                      
        variable_name : str
            Name of a variable.
                      
        x_colorbar : float
            Position of a colorbar on axis X.

        Returns
        -------
        Plotly Scattergl object
            X- and Y-coordinates of a Scattergl object are not define. The object contains only a colorbar. 
        """

        return go.Scattergl(x=[None], y=[None], mode='markers', showlegend=False, legendgroup=variable_name,
                            marker=dict(colorscale=colorset, cmin=valueset[0], cmax=valueset[-1]+tick_interval,
                                        colorbar=dict(tickvals=valueset, outlinecolor='black', outlinewidth=0.5, x=x_colorbar)))
    
    @classmethod
    def make_colorbar_continuous(cls, data, colormap):
        """
        Creates a continouos colorbar.
    
        Parameters
        ----------
        data : xarray
            Values of a variable and a range of other parameters describing a variable and a grid.
                 
        colormap : str
            Name of a colormap.
            
        Returns
        -------
        Plotly Scattergl object
            X- and Y-coordinates of a Scattergl object are not define. The object contains only a colorbar.
        """      
        
        return go.Scattergl(x=[None],y=[None], mode='markers', showlegend=False, 
                            marker=dict(colorscale=colormap, showscale=True, cmin=np.min(data), cmax=np.max(data), 
                                        colorbar=dict(outlinecolor='black', outlinewidth=0.5)), hoverinfo='none')
    
    @classmethod
    def set_camera_eye(cls, lons, lats):
        """
        Defines a camera eye. 
    
        Parameters
        ----------
        lons, lats : Numpy array
            Numpy arrays contain longitudes or latitudes in radians.
            
        Returns
        -------
        Numpy array
            Normal vector at a center point of a region.  
        """        
        
        # define two surfaces, one above other 
        cartesian_coordinates = cls.spherical_to_cartesian(lons, lats, radius=6371229.0)
        xs = cartesian_coordinates.T[0]
        ys = cartesian_coordinates.T[1]
        zs = cartesian_coordinates.T[2]
        
        cartesian_coordinates_above = cls.spherical_to_cartesian(lons, lats, radius=6371230.0)
        xs_above = cartesian_coordinates_above.T[0]
        ys_above = cartesian_coordinates_above.T[1]
        zs_above = cartesian_coordinates_above.T[2]
        
        # define central points of surfaces
        x_mean = np.mean(xs)
        y_mean = np.mean(ys)
        z_mean = np.mean(zs)
        center_point = np.array([x_mean, y_mean, z_mean])
        
        x_mean_above = np.mean(xs_above)
        y_mean_above = np.mean(ys_above)
        z_mean_above = np.mean(zs_above)
        center_point_above = np.array([x_mean_above, y_mean_above, z_mean_above])
        
        # calculate normal vector at a central point 
        normal = center_point_above - center_point

        return normal


from pathlib import Path
from typing import Union, List
import glob


def clean_paths(paths: Union[Path, List[Path], List[str], str],
                check_files_exist: bool = True
                ,) -> List[Path]:
    """
    Take the paths in all the usual forms and convert them to a list of Path objects.
    Parameters
    ----------
    paths

    Returns
    -------

    """

    # Check arguments


    if not isinstance(paths, (list, str, tuple, Path)):
        raise NotImplementedError("Unsupported type of argument: %s" % type(paths))    

    if not isinstance(paths, list):
        paths = [paths]

    # Expand paths using glob
    expanded_paths = [glob.glob(str(fp)) for fp in paths]

    # Merge elements in a flat list
    expanded_paths = [item for sublist in expanded_paths for item in sublist]

    # Convert the elements to Path and expand
    path_objects = [Path(fp).resolve() for fp in expanded_paths]
    if check_files_exist:
        for fp in path_objects:
            assert fp.exists()
    return path_objects



from .backend import InteractiveBackend

import numpy as np

import matplotlib.pyplot as plt
from bokeh.models import FixedTicker, LinearColorMapper, ColorBar



class InteractiveBackendBokeh(InteractiveBackend):
    """
    Implementation of the interactive plotting backend for Bokeh.
    """
    
    @classmethod
    def make_colorbar(cls, colorset, valueset, tick_interval, variable_name, x_position=None):
        """
        Creates a colorbar for Bokeh.

        Parameters
        ----------
        colorset : list of tuples
            Set of ordered unique colors. List containing tuples mapping a normalized value to a HEX or RGB. 
                 
        valueset : Numpy array
            Set of ordered unique double values.
                 
        tick_interval : float
            Value of intervals between ticks in a colorbar.
                      
        variable_name : str
            Name of a variable.
                      
        x_position : float, optional, default=None
            Position of a colorbar on axis X.

        Returns
        -------
        Bokeh ColorBar object
        """
        
        # take only a colors definition without normalization
        colorset_modified = []
        for i in range(len(colorset)):
            colorset_modified.append(colorset[i][1])
            
        # define colorbar parameters    
        ticker = FixedTicker(ticks=valueset)
        color_mapper = LinearColorMapper(palette=colorset_modified, low=valueset[0], high=valueset[-1]+tick_interval)

        return ColorBar(color_mapper=color_mapper, ticker=ticker, label_standoff=12,
                        border_line_color=None, location=(0, 0), name=variable_name)

    @classmethod
    def separate_streamline_data(cls, streamlines_path):
        """
        Separates individual streamlines and also separates X- and Y- coordinates of them.

        Parameters
        ----------
        streamlines_path : array
            3-dimensional array contains X- and Y-coordinates of streamlines.
                  
        Returns
        -------
        xs, ys : list of lists
            Lists of lists contain X- or Y-coordinates of streamlines.
        """
        
        n = 0; k0 = 0
        k = []; xs = []; ys = []   
        
        # find beginning and ending indices for individual streamlines 
        for i in range(1,len(streamlines_path)):
            if streamlines_path[i-1,1,1] != streamlines_path[i,0,1] or streamlines_path[i-1,1,0] != streamlines_path[i,0,0]:
                n += 1
                k.append(i-1)
                
        # separate X- and Y- coordinates
        for ks in k:
            k1 = ks
            xsi = streamlines_path[k0:k1,1,0]
            ysi = streamlines_path[k0:k1,1,1]
            xs.append(xsi)
            ys.append(ysi)
            k0 = ks+1
    
        plt.close()
    
        return xs, ys
    
    @classmethod
    def separate_vector_data(cls, lons, lats, us, vs, scale, ref_arrow, lon0_map, lat1_map, lat1_map_arrow):
        """
        Separates parameters of vectors and a reference arrow. Creates arrowheads for vectors.

        Parameters
        ----------
        lons, lats : Numpy array
            2-dimensional arrays contain X- or Y-coordinates of a reduced number of grid cells.

        us, vs : Numpy array
            2-dimensional arrays contain values of horizontal or vertical components of a variable.
             
        scale : float, optional, default=0.02
            Possibility to scale a size of arrows(ideally to avoid overlap).
            
        ref_arrow : int
            Value of a reference arrow.
            
        lon0_map, lat1_map : float
            Longitude and latitude from region borders (left top corner) for definition of a reference arrow position.
            
        lat1_map_arrow : float
            Value equals approximately 10% of a height of a region. Used for definition of a reference arrow position for 2D plots.

        Returns
        -------
        xs, ys : list of arrays
            Lists of arrays contain X- or Y-coordinates of vectors and arrowheads.
                  
        xs_ref_arrow, ys_ref_arrow : list of arrays
            Lists of arrays contain X- or Y-coordinates of a reference arrow.
        """ 
        
        # define position of the reference arrow in Y-axis
        lat1_map = lat1_map - 0.8*lat1_map_arrow

        # define parameters of vectors
        speed = np.sqrt(us*us + vs*vs)
        theta = np.arctan(abs(vs/us))
        
        # define parameters for plotting
        x0 = lons.flatten()
        y0 = lats.flatten()
        
        length = speed.flatten()*scale
        angle = theta.flatten() 
        x1 = x0 + np.sign(us).flatten() * length * np.cos(angle)
        y1 = y0 + np.sign(vs).flatten() * length * np.sin(angle)
        
        # define arrowhead parameters
        x1_head = x1 - np.sign(us).flatten() * length/5 * np.cos(angle + np.radians(30))
        y1_head = y1 - np.sign(vs).flatten() * length/5 * np.sin(angle + np.radians(30))
        x2_head = x1 - np.sign(us).flatten() * length/5 * np.cos(angle - np.radians(30))
        y2_head = y1 - np.sign(vs).flatten() * length/5 * np.sin(angle - np.radians(30))
        
        # separate X- and Y-coordinates of vectors
        xs = []; ys = []  
        for i in range(len(x0)):
            xsi = np.asarray([x0[i], x1[i]])
            ysi = np.asarray([y0[i], y1[i]])
            xs.append(xsi)
            ys.append(ysi)
        # add arrowheads to a vector definition    
        for i in range(len(x0)):
            xsi_head = np.asarray([x1_head[i], x1[i], x2_head[i]])
            ysi_head = np.asarray([y1_head[i], y1[i], y2_head[i]])
            xs.append(xsi_head)
            ys.append(ysi_head)
        
        # define and separate reference arrow coordinates
        ref_arrow_length  = ref_arrow*scale
        
        xs_arrow_line = np.asarray([lon0_map+0.5, lon0_map+0.5+ref_arrow_length])
        ys_arrow_line = np.asarray([lat1_map, lat1_map])
            
        x1_arrow_head = xs_arrow_line[1] - ref_arrow_length/5 * np.cos(np.radians(30))
        y1_arrow_head = ys_arrow_line[1] - ref_arrow_length/5 * np.sin(np.radians(30))
        x2_arrow_head = xs_arrow_line[1] - ref_arrow_length/5 * np.cos(-np.radians(30))
        y2_arrow_head = ys_arrow_line[1] - ref_arrow_length/5 * np.sin(-np.radians(30))
            
        xs_arrow_head = np.asarray([x1_arrow_head, xs_arrow_line[1], x2_arrow_head])
        ys_arrow_head = np.asarray([y1_arrow_head, ys_arrow_line[1], y2_arrow_head])
            
        xs_ref_arrow = []; ys_ref_arrow = []
        xs_ref_arrow.append(xs_arrow_line); xs_ref_arrow.append(xs_arrow_head)
        ys_ref_arrow.append(ys_arrow_line); ys_ref_arrow.append(ys_arrow_head)
        
        return xs, ys, xs_ref_arrow, ys_ref_arrow


import enstools.clustering
import enstools.core
import enstools.filters
import enstools.interpolation
import enstools.io
import enstools.misc
import enstools.opendata
import enstools.plot
import enstools.post
import enstools.scores



import xarray


def metric(reference: xarray.DataArray, target: xarray.DataArray,
           extra_argument: str = "Ideally extra arguments must have a default value") -> xarray.DataArray:
    """
    Description: A meaningful description of the metric with references if necessary would be nice.

    Functions
    :param reference:
    :param target:
    :param extra_argument:
    :return:
    """
    print(extra_argument)
    return target - reference



"""
Routines for Evaluation of model forecasts
"""
# Imports for typing
from typing import Callable
from xarray import DataArray

# Import actual scores
from .DisplacementAmplitudeScore import das
from .continuous_ranked_probability_score import continuous_ranked_probability_score
from .kolmogorov_smirnov import kolmogorov_smirnov, kolmogorov_smirnov_index
from .kolmogorov_smirnov_multicell import kolmogorov_smirnov_multicell
from .normalized_root_mean_square_error import (mean_square_error,
                                                root_mean_square_error,
                                                normalized_root_mean_square_error,
                                                normalized_root_mean_square_error_index,
                                                )
from .pearson_correlation import pearson_correlation, pearson_correlation_index
from .structural_similarity_index import structural_similarity_index, structural_similarity_log_index
from .peak_signal_to_noise_ratio import peak_signal_to_noise_ratio
from .positivity import positivity

# Add some aliases
ssim_I = structural_similarity_log_index
correlation_I = pearson_correlation_index
nrmse_I = normalized_root_mean_square_error_index
ks_I = kolmogorov_smirnov_index
psnr = peak_signal_to_noise_ratio


def add_score_from_file(file_path: str):
    """
    Read a python file and add the function with the same name as the file to the available scores in enstools.scores
    """

    import importlib
    import os
    module_name = os.path.basename(file_path).strip(".py")

    import importlib.util
    import sys

    # Get spec from the file path
    spec = importlib.util.spec_from_file_location(module_name, file_path)

    # Get the module
    mod = importlib.util.module_from_spec(spec)

    # Load the module
    spec.loader.exec_module(mod)

    # Get the function
    function = mod.__getattribute__(module_name)

    register_score(function, module_name)


def register_score(function: Callable[[DataArray, DataArray], DataArray], name: str):
    """
    Register a given function as a score.
    """
    import sys
    # Add the function as a new score in module enstools.scores
    setattr(sys.modules[__name__], name, function)



from enstools.core import getstatusoutput
import logging
import re
import os


def get_file_type(filename: str, only_extension=False):
    """
    use the first bytes to guess the type of input files

    Parameters
    ----------
    filename : string
            name of the file to check

    only_extension : bool
            if True, only the file name is checked. This is useful for not yet created files.

    Returns
    -------
    string
            NC, HDF, GRIB
    """
    # guess the type based on extension
    file_type_based_on_extension = None
    file_type = None
    basename, extension = os.path.splitext(filename)
    if extension in [".nc", ".nc3", ".nc4"]:
        file_type_based_on_extension = "NC"
    elif extension in [".grb", ".grib", ".grb2", ".grib2"]:
        file_type_based_on_extension = "GRIB"

    if not only_extension:
        # is the file present at all?
        if not os.path.exists(filename):
            raise IOError("file '%s' not found!" % filename)

        # read the first bytes and decide based on the content
        with open(filename, "rb") as f:
            first_bytes = f.read(12)

            # netcdf3 file
            if first_bytes.startswith(b"CDF"):
                file_type = "NC"
            # netcdf4
            elif b"HDF" in first_bytes[0:4]:
                file_type = "HDF"
            # grib
            elif b"GRIB" in first_bytes:
                file_type = "GRIB"

    # return the best knowledge of the file type
    if file_type is not None:
        return file_type
    else:
        return file_type_based_on_extension




from .backend import InteractiveBackend

import numpy as np

import matplotlib.pyplot as plt
import plotly.graph_objs as go



class InteractiveBackendPlotly(InteractiveBackend):
    """
    Implementation of the interactive plotting backend for Plotly.
    """
    
    @classmethod
    def make_scattergl_map(cls, xs, ys):
        """
        Makes a Plotly Scattergl object for coastlines and country borders.

        Parameters
        ----------
        xs, ys : list of lists
            Lists of lists contain separated X- or Y-coordinates of paths.

        Returns
        -------
        list
            List contains a Plotly Scattergl objects. 
        """
        
        traces = []
        for i in range(len(xs)):
            traces.append(go.Scattergl(x=xs[i], y=ys[i], mode='lines', line=go.scattergl.Line(color="black", width=0.7),
                                       name='map', showlegend=False, opacity=0.7))
        return traces
    
    @classmethod
    def make_filled_contour(cls, x, y, z, variable_name, colormap, x_colorbar):
        """
        Creates filled contours using specialized Plotly function.

        Parameters
        ----------
        x, y : array
            Arrays contain X- or Y-coordinates of grid cells.
                      
        z : array
            2-dimensional array contain values of a variable for every grid cell.
                      
        variable_name : str
            Name of a variable.
                       
        colormap : str
            Name of a colormap.
                      
        x_colorbar : float
            Position of a colorbar on axis X.

        Returns
        -------
        Plotly Contour object
            Filled contours with a colorbar. 
        """

        return go.Contour(z=z, x=x, y=y, showlegend=True, name=variable_name,
                          autocontour=True, ncontours=10, showscale=True, colorscale=colormap,
                          colorbar=dict(outlinecolor='black', outlinewidth=0.5, bordercolor='white', x=x_colorbar))

    @classmethod
    def make_unfilled_contour(cls, xs, ys, color, value, variable_name, showlegend=True):
        """
        Creates a Scattergl object for an unfilled contour.

        Parameters
        ----------
        xs, ys : list
            Lists contain X- or Y-coordinates of a contour.
                      
        color : str
            Color of a contour.
                       
        value : str
            Value of a contour.
                      
        variable_name : str
            Name of a variable.
                       
        showlegend : bool, default=True
            Definition of displaying of a contour at a legend.

        Returns
        -------
        Plotly Scattergl object
            Scattergl object contains an unfilled contour.
        """ 

        return go.Scattergl(x=xs, y=ys, mode='lines', line=go.scattergl.Line(color=color, width=1),
                            showlegend=showlegend, legendgroup=variable_name, name=value)
    
    @classmethod
    def make_colorbar(cls, colorset, valueset, tick_interval, variable_name, x_colorbar):
        """
        Creates a colorbar for Plotly.

        Parameters
        ----------
        colorset : list of tuples
            Set of ordered unique colors. List containing tuples mapping a normalized value to a HEX or RGB. 
                 
        valueset : Numpy array
            Set of ordered unique double values.
                 
        tick_interval : double
            Value of intervals between ticks in a colorbar.
                      
        variable_name : str
            Name of a variable.
                      
        x_colorbar : float
            Position of a colorbar on axis X.

        Returns
        -------
        Plotly Scattergl object
            X- and Y-coordinates of a Scattergl object are not define. The object contains only a colorbar. 
        """

        return go.Scattergl(x=[None], y=[None], mode='markers', showlegend=False, legendgroup=variable_name,
                            marker=dict(colorscale=colorset, cmin=valueset[0], cmax=valueset[-1]+tick_interval,
                                        colorbar=dict(tickvals=valueset, outlinecolor='black', outlinewidth=0.5, x=x_colorbar)))
    
    @classmethod 
    def make_scattergl_streamline(cls, streamlines_path, color, showlegend=True):
        """
        Makes a Scattergl object for streamlines.

        Parameters
        ----------
        streamlines_path : array
            3-dimensional array contains X- and Y-coordinates of streamlines.
            
        color : str
            Defines a color of streamlines.
        
        showlegend : bool, optional, default=True
            Definition of displaying of streamlines at a legend.

        Returns
        -------
        Plotly Scattergl object                           
        """

        return go.Scattergl(x=streamlines_path.T[0], y=streamlines_path.T[1],
                            mode='lines', line=go.scattergl.Line(color=color, width=1), 
                            showlegend=showlegend, legendgroup='streamlines', name='streamlines')
   